﻿///
/// 
/// Sample Mission
/// 
/// @Author    Team Fusion - Artist
/// @Version   1.0.10
/// 
/// === History ===
/// ** 1.0.10 **
/// - chg: CAiMissionCommand updated to 1.0.1: set limits per army
/// - chg: CMissionPlayers updated to 1.0.1
/// ** 1.0.9 **
/// - chg: replaced deprecated sPathMyself with this.PathMyself
/// - chg: CRadar updated to 1.0.3 - also using PathMyself now
/// ** 1.0.8 **
/// - chg: Moved some module instantiating into constructor Mission()
/// - chg: CRearmRefuel updated to 1.0.3 (minimum fuel)
/// - chg: CChatCommands updated to 1.0.4 (dynamic commands)
/// /// ** 1.0.7 **
/// - new: CTargetAreas - geographical areas as bombing targets (similiar to CTargetAirfield)
/// - chg: CTargetAirfields updated to 1.0.5
/// ** 1.0.6 **
/// - new: CTargetGroups - by Matkenhauser
/// ** 1.0.5 **
/// - chg: CRadar updated to 1.0.2 - Fixed a bug
/// ** 1.0.4 **
/// - chg: CRadar updated to 1.0.1 - Fixed a bug
/// ** 1.0.3 **
/// - new: CMissionPlayers (see all usage of m_MissionPlayers)
/// - chg: CAiMissionCommand: Option to prevent AiMissions orderable beyond specified player number
///        see: m_AiMissionCommand.SetMaxPlayers(30, m_MissionPlayers);
/// - new: CRadar (see all usage of m_Radar)  
/// - new: CSectionFile-v1.0.0.cs (reading of .mis and other section files)

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;          /// needed to end mission by calling Process.GetCurrentProcess().Kill()
using System.Text;
using maddox.game;                 /// AMission
using maddox.game.world;           /// AiActor
using maddox.GP;

/// Include additional .cs files
/// Syntax: //$include [drive | path | $user]\path\scriptfilename.cs
///     "$user"       expands to WindowsDrive:\User\<Username>\Documents\1C SoftClub\il-2 sturmovik cliffs of dover
///     "C:\path\"    is absolute
///     ".\" or "..\" are relative to this scripts location
//$include .\modules\CAiMissionCommand-v1.0.1.cs
//$include .\modules\CArmies-v1.0.1.cs
//$include .\modules\CBroadcast-v1.0.1.cs
//$include .\modules\CChatCommands-v1.0.4.cs
//$include .\modules\CKillDisusedPlanes-v1.0.1.cs
//$include .\modules\CMissionMenu-v1.0.0.cs
using MissionMenu;
//$include .\modules\CMissionStatus-v1.0.1.cs
//$include .\modules\CMissionPlayers-v1.0.1.cs
//$include .\modules\CRadar-v1.0.3.cs
//$include .\modules\CRearmRefuel-v1.0.3.cs
//$include .\modules\CTargetAirfields-v1.0.5.cs
//$include .\modules\CTargetAreas-v1.0.0.cs
//$include .\modules\CTargetConvoys-v1.0.0.cs
//$include .\modules\CTargetDamage-v1.0.1.cs
//$include .\modules\CTargetGroups-v1.0.1.cs

public class Mission : AMission {
    protected string m_sUserDoc = null;
    protected string m_sMyFolder = null;
    /// Helper Modules
    protected CAiMissionCommand  m_AiMissionCommand;
    protected CBroadcast         m_Broadcast;
    protected CChatCommands      m_ChatCmd;
    protected CMissionMenu       m_MissionMenu;
    protected CMissionPlayers    m_MissionPlayers;
    protected CMissionStatus     m_MissionStatus;
    protected CKillDisusedPlanes m_KillDisusedPlanes;
    protected CRadar             m_Radar;
    protected CRearmRefuel       m_RearmRefuel;
    protected CTargetAirfields   m_TargetAirfields;
    protected CTargetAreas       m_TargetAreas;
    protected CTargetConvoys     m_TargetConvoys;
    protected CTargetGroups      m_TargetGroups;
    /// ServerStat System
    protected AMission m_ServerStatMis = null;

    /// <summary>
    /// 
    /// </summary>
    public Mission()
    {
        m_MissionPlayers = new CMissionPlayers(this);
    }
    /// <summary>
    /// Event: Starting the mission
    /// </summary>
    public override void OnBattleStarted()
    {
        MissionNumberListener = -1;
        base.OnBattleStarted();

        m_sUserDoc  = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + @"\1C SoftClub\il-2 sturmovik cliffs of dover\";
        m_sMyFolder = Path.GetDirectoryName(this.PathMyself)+@"\";


        /// ***************************
        /// PART 0: INSTANTIATE
        /// ***************************
        /// PART 1: HELPER MODULES
        m_MissionStatus = new CMissionStatus();
        m_Broadcast = new CBroadcast(this, Time);
        m_ChatCmd = new CChatCommands(this);
        m_AiMissionCommand = new CAiMissionCommand(this, m_Broadcast);
        m_Radar = new CRadar(this, m_MissionPlayers);
        m_MissionMenu = new MissionMenu.CMissionMenu(this);
        m_RearmRefuel = new CRearmRefuel(this);
        /// Damage and kill abandoned aircraft
        /// This is used for making sure that the plane a player has left
        /// cannot be regained (outside views) and is removed from the game.
        /// see OnPlaceLeave()
        m_KillDisusedPlanes = new CKillDisusedPlanes(this);
        /// PART 2: MISSION TARGETS
        m_TargetAirfields = new CTargetAirfields(this);
        m_TargetAreas = new CTargetAreas(this);
        m_TargetConvoys = new CTargetConvoys(this);
        m_TargetGroups = new CTargetGroups(this);


        /// ***************************
        /// PART 1: HELPER MODULES
        /// ***************************
        /// === PLAYERS ===

        /// === MISSION STATUS ===
        /// Mission Status: score, time left, 
        /// Set mission time X hours, Y minutes
        m_MissionStatus.SetMissionDuration(2, 30);
        /// Set X points to win mission
        m_MissionStatus.SetPointsToWin(2);

        /// === BROADCAST SYSTEM ===
        /// see OnTickGame(), OnBombExplosion(), and EndMission()
        /// Print mission's status and time left in stated intervals (minutes)
        /// Text is composed in this mission class' methods we pass to broadcast here
        m_Broadcast.AddAutoMessageChat(CArmies.ARMY_ALL, 5, this.GetScoreStringForChat);
        m_Broadcast.AddAutoMessageScreen(CArmies.ARMY_ALL, 30, this.GetScoreStringForScreen);

        /// === CHAT COMMANDS ===
        /// initialize and configure chat commands like <![CDATA["<tl"]]>
        /// Texts are composed in this mission class' methods we pass to ChatCommands here
        /// The help texts are shown when user types <![CDATA["<help"]]> in chat
        m_ChatCmd.AddCommand("<tl",       this.GetTimeLeftString, "<tl => Get time left in mission");
        m_ChatCmd.AddCommand("<obj<blue", this.GetScoreStringBlue, "<obj<blue => Get achieved blue objectives");
        m_ChatCmd.AddCommand("<obj<red",  this.GetScoreStringRed, "<obj<red => Get achieved red objectives");

        /// === ADD AI MISSIONS ===
        /// The 4. Mission Menu is autofilled (see further down)
        /// Mission Timeout in seconds - time after the next mission can be ordered
        m_AiMissionCommand.SetMissionTimeout(900);
        /// Limits: 
        /// Set a max player (all armies) limit beyond which no missions can be ordered (0 for no limit)
        m_AiMissionCommand.SetMaxPlayers(30, m_MissionPlayers);
        /// Set a max player (player's own army) limit beyond which no missions can be ordered (0 for no limit)
        /// This overrides SetMaxPlayers: if max players is 30 and there are 25 blue players and 10 red players and
        /// you set this to 15 then red will be allowed to start a mission.
        m_AiMissionCommand.SetMaxPlayersOwnArmy(15, m_MissionPlayers);
        /// Set a maximum of concurrent missions (all armies)
        m_AiMissionCommand.SetMaxConcurrentMissions(1);
        /// Set a maximum of concurrent missions per army
        /// This overrides SetMaxConcurrentMissions: if red has 2 missions running and you set this
        /// to 2 then blue will still be allowed to start 2 missions.
        m_AiMissionCommand.SetMaxConcurrentMissionsPerArmy(1);

        /// Add orderable missions
        m_AiMissionCommand.AddMission(CArmies.ARMY_RED, m_sMyFolder + @"aimissions\AiRed_Attack_Amseat_No1.mis",
            /// mission menu text
            "1 Wellington, angels 3, start NE of Amseat No1; target: Amseat No1",
            /// friendly announcement
            "Attention: 1 Wellington will start NE of Amseat No1, angels 3; target: Amseat No1",
            /// friendly intel
            "1 Wellington, angels 3, started NE of Amseat No1; target: Amseat No1",
            /// enemy intel
            "Attack on Amseat No1 expected from NE"
        );
        m_AiMissionCommand.AddMission(CArmies.ARMY_RED, m_sMyFolder + @"aimissions\AiRed_Attack_BlueCampA.mis",
            "3 Wellingtons, angels 1 , start 3 minutes E of blue camp A; target: blue camp A",
            "Attention: 3 Wellingtons will start 3 minutes E of blue camp A, angels 1; target: blue camp A",
            "3 Wellingtons, angels 1, started 3 minutes E of blue camp A; target: blue camp A",
            "Attack on blue camp A expected from E"
        );
        m_AiMissionCommand.AddMission(CArmies.ARMY_RED, m_sMyFolder + @"aimissions\AiRed_Attack_BlueRadar.mis",
            "3 Wellingtons, angels 1 , start 3 minutes E of blue radar; target: blue radar",
            "Attention: 3 Wellingtons will start 3 minutes E of blue radar, angels 1; target: blue radar",
            "3 Wellingtons, angels 1, started 3 minutes E of blue radar; target: blue radar",
            "Attack on blue radar expected from E"
        );
        m_AiMissionCommand.AddMission(CArmies.ARMY_RED, m_sMyFolder + @"aimissions\AiRed_Attack_FortCapuzzo.mis",
            "3 Wellingtons, angels 3 , start 3 minutes NE of Fort Capuzzo; target: Car Pool at Fort Capuzzo",
            "Attention: 3 Wellingtons will start 3 minutes NE of Fort Capuzzo, angels 3; target: Car Pool at Fort Capuzzo",
            "3 Wellingtons, angels 3, started 3 minutes NE of Fort Capuzzo; target: Car Pool at Fort Capuzzo",
            "Attack on Car Pool at Fort Capuzzo expected from NE"
        );

        m_AiMissionCommand.AddMission(CArmies.ARMY_BLUE, m_sMyFolder + @"aimissions\AiBlue_Attack_Buq_Buq.mis",
            "1 He-111 at 1000m, start NW of Buq Buq: Target: Buq Buq",
            "Achtung: 1 He-111 start NW of Buq Buq at 1000m: Target: Buq Buq",
            "1 He-111 at 1000m, started NW of Buq Buq: Target: Buq Buq",
            "Attack on Buq Buq expected from NW"
        );
        m_AiMissionCommand.AddMission(CArmies.ARMY_BLUE, m_sMyFolder + @"aimissions\AiBlue_Attack_RedCampA.mis",
            "3 He-111 at 500m attack red camp A",
            "Achtung: 3 He-111 start 3 min SW red camp A, target: red camp A",
            "3 He-111 at 500m, started 3 min SW red camp A, target: red camp A",
            "Attack on red camp A expected from SW"
        );
        m_AiMissionCommand.AddMission(CArmies.ARMY_BLUE, m_sMyFolder + @"aimissions\AiBlue_Attack_RedRadar.mis",
            "3 He-111 at 500m attack red radar station",
            "Achtung: 3 He-111 start 3 min W red radar station, target: radar station",
            "3 He-111 at 500m, started 3 min W red radar station, target: radar station",
            "Attack on radar expected from W"
        );

        /// === RADAR ===
        /// The 4. Mission Menu is autofilled (see further down)
        /// Time (seconds) between request and report
        m_Radar.SetResponseTime(3);
        /// Minimum time (seconds) between two requests
        m_Radar.SetMinRequestInteval(30);
        /// Only aircraft above this height will be reported
        m_Radar.SetMinObservableHeight(400.0);
        /// Radius within which airplanes are considered to be within one 'pulk'.
        m_Radar.SetAircraftPulkRadius(1400.0);
        /// Necessary minimum number of aircraft in a "pulk" to make it reported
        m_Radar.SetMinAircraftInPulk(1);

        /// Add headquarters and observer stations to radar manager
        /// 
        /// A headquarter is a collection of one or more observers.
        /// 
        /// Each headquarter and each observer station must have a "Static### ..." entry in the .mis file in the [Stationary] section.
        /// 
        /// Place a suitable object from "Static" (not building!) in the FMB then 'mouse over' that spot to find the x and y coords in
        /// the status bar at the bottom of the screen. Then open the .mis file and locate a "Static..." listed with the same coordinates
        /// with in the [Stationary] section. This will give you the "Static###" to use here.
        /// Hint 1: You can also (a) make a backup copy, (b) add the radar objects, and finally (c) make a diff/compare of the two files.
        /// Hint 2: We recommend that you place the radar headquarters and stations *before* any other stationary objects. Makes life easier.
        /// 
        /// There is *no* need to create triggers for the objects. If you configure a headquarter/observer as destroyable, CRadar will
        /// automatically generate (and later handle) the destruction trigger that makes the object destroyable by bombing.
        /// (see OnTrigger below, too)
        /// 
        /// The points of a headquarter are awarded when all observers have been destroyed. You can make a _headquarter_ destroyable by itself.
        /// 
        /// Create a blue radar headquarter
        CRadar.CRadarHeadquarter RadarRedHq1 = m_Radar.AddHeadquarter(
            "Static0",        /// The static object representing the Headquarter
            CArmies.ARMY_RED, /// if you use ARMY_NONE, owning army is determined by frontline/position
            "NameIfBlue",     /// Name if blue (fronts may shift)
            "NameIfRed",      /// Name if red  (fronts may shift)
            "CallSign24",     /// Callsign (also audio file)
            100000.0,         /// range (for pilot to use this Headquarter)
            1,                /// award this number of points for destruction of headquarter or all its observers
            true              /// is this destroyable by itself
        );
        /// add an observer (you can add as many as you want)
        RadarRedHq1.AddObserver(
            "Static1",      /// The static object representing the observer/radar
            "ObserverName", /// Name
            60000.0,        /// radius of area covered around the observer station
            true            /// is this destroyable
        );

        /// Create a blue radar
        CRadar.CRadarHeadquarter RadarBlueHq1 = m_Radar.AddHeadquarter("Static2", CArmies.ARMY_BLUE, "NameIfBlue", "NameIfRed", "CallSign14", 100000.0, 1, true);
        RadarBlueHq1.AddObserver("Static3", "ObserverName2", 60000.0, true);

        /// activate radar manager (*must* be last, after all HQs and observers are added).
        m_Radar.Activate();

        
        /// === POPULATE THE "4. Mission" MENU ===
        /// 4. Mission
        ///    x. Bomber Command
        ///       1. Get Status Report
        ///       2. Send Bombers
        ///          x. dynamic commands created by m_AiMissionCommand.OnMissionMenuMakeCommands()
        ///    x. Rearm/Refuel
        ///       ... auto filled by CRearmRefuel
        /// Add the Bomber Command Menu - also an example how to show different texts for red and blue players
        MissionMenu.MenuItem menBmbCmd = m_MissionMenu.AddMenu(new string[] { "Bomber Command", "Kampfgeschwader Orders" });
        /// Add the "Get report" - also an Example how to add a command with a 'callback' function
        menBmbCmd.AddCommand("Get Status Report", m_AiMissionCommand.OnMissionMenuGetReport);
        /// Add the (dynamic!) "Send bomber" sub menu - also an example how to not defining a static menu at start, but only when it's actually shown
        menBmbCmd.AddDynamicMenu("Send Bombers", m_AiMissionCommand.OnMissionMenuMakeCommands, m_AiMissionCommand.OnMissionMenuSelected);

        /// Add a Menu for Radar
        MissionMenu.MenuItem menRadar = m_MissionMenu.AddMenu(new string[] { "Radar Ground Control", "Freya Radar Control" });
        menRadar.AddDynamicMenu(new string[] { "Select Dispatch Controller", "Select Leitstelle" }, m_Radar.OnMissionMenuSelectHQ_Make, m_Radar.OnMissionMenuSelectHQ_Select);
        menRadar.AddCommand("Get Enemy Contacts", m_Radar.OnMissionMenuGetEnemyContacts);
        menRadar.AddCommand("Get Friendly Contacts", m_Radar.OnMissionMenuGetFriendlyContacts);

        /// Add a menu for Rearm/Refuel (its commands are populated by CRearmRefuel, see further down)
        MissionMenu.MenuItem menRnR = m_MissionMenu.AddMenu("Rearm / Refuel");

        /// === REARM / REFUEL ===
        /// Add Rearm/Refuel command handler(s) to ChatCommands, 'help' shows options
        m_RearmRefuel.AddToChatCommands(m_ChatCmd);
        /// Add Rearm/Refuel command handler(s) to Mission menu (int[] lists percentage options)
        m_RearmRefuel.AddToMissionMenu(menRnR, new int[] { 5, 10, 25, 50, 75, 100 });
        /// See the CRearmRefuel.Set... functions for details and additional configurations
        m_RearmRefuel.SetMinimumFuel(true);
        m_RearmRefuel.SetRearmDuration(60);
        m_RearmRefuel.SetRefuelDuration(80);
        m_RearmRefuel.SetMinimumRefuelDuration(60);

        /// === SERVER STATS ===
        /// Initialize ServerStat System
        /// This sample mission has it disabled by default. Check the modules\ServerStats\CServerStats.ini
        GamePlay.gpPostMissionLoad(m_sMyFolder + @"modules\ServerStats\CServerStats-v1.0.0-Mission.mis");


        /// ***************************
        /// PART 2: MISSION TARGETS
        /// ***************************

        /// === Type Type Airfield ===

        /// Setup Airfields as Target
        /// see OnBombExplosion()
        /// sName
        ///     Name of Airfield
        ///     remember that airfield names are not what you see on the map, but
        ///     the 'internal' names, see FMB Mission Objects.
        /// sBirtplace
        ///     Name of Birthplace
        ///     The exact string the you named the birthplace with in the FMB
        ///     (see the .mis file, section [BirthPlace])
        /// iKgToDamage100
        ///     Set amount of kilogram explosives needed for 100% damage
        ///     He 111 (H-2, P), Ju 88 A-1, Wellington: 2000kg
        ///     100% of 9000 means 5 He 111s or 5 Wellington
        /// iPoints
        ///     Points to award for destruction of this airfield

        /// a red airfields as target
        m_TargetAirfields.Add(CArmies.ARMY_RED, "Buq_Buq_LG01", "Buq Buq", 9000, 1);
        /// a blue airfields as target
        m_TargetAirfields.Add(CArmies.ARMY_BLUE, "Amseat_No1", "Amseat No1", 9000, 1);
        /// Percentage of destruction assumed to have 'killed' an airfield
        /// With our values from above that means:
        /// 100% = 9000kg, 50% = 4500kg, 10% = 900kg
        m_TargetAirfields.SetDestroyedAtPercent(100);

        /// === Target Type Type Area ===
        
        /// Setup areas as targets
        /// see OnBombExplosion()
        /// sName
        ///     Name of the area
        /// dPosX
        ///     X Position of areas center
        /// dPosY
        ///     Y Position of areas center
        /// dRadius
        ///     size of area around center position in meters
        /// iKgToDamage100
        ///     Set amount of kilogram explosives needed for 100% damage
        ///     He 111 (H-2, P), Ju 88 A-1, Wellington: 2000kg
        ///     100% of 9000 means 5 He 111s or 5 Wellington
        /// iPoints
        ///     Points to award for destruction of this airfield

        /// a blue car pool as area target
        m_TargetAreas.Add(CArmies.ARMY_BLUE, "Car Pool Fort Capuzzo", 264627, 130798, 300, 9000, 1);
        /// Percentage of destruction assumed to have 'killed' an area
        /// With our values from above that means:
        /// 100% = 2000kg, 50% = 1000kg, 10% = 200kg
        m_TargetAreas.SetDestroyedAtPercent(100);

        /// === Target Type Convoy ===

        /// Setup Convoys as Target
        /// see OnTrigger()
        /// one convoy: blue
        CTargetConvoy BlueConvoy1 = new CTargetConvoy(this, CArmies.ARMY_BLUE, "Blue Tanker Convoy", 3);
        BlueConvoy1.AddUnitByTriggerName("GB_ship1");
        BlueConvoy1.AddUnitByTriggerName("GB_ship2");
        m_TargetConvoys.AddConvoy(BlueConvoy1);

        /// one convoy: red
        CTargetConvoy RedConvoy1 = new CTargetConvoy(this, CArmies.ARMY_RED, "Red Tanker Convoy", 3);
        RedConvoy1.AddUnitByTriggerName("GB_ship3");
        RedConvoy1.AddUnitByTriggerName("GB_ship4");
        m_TargetConvoys.AddConvoy(RedConvoy1);

        /// === Target Type (homogenous) Groups ===
        
        /// Setup groups as targets
        /// see OnTrigger()
        /// a blue group
        CTargetGroup BlueCampA = new CTargetGroup(this, CArmies.ARMY_BLUE, "Blue Camp A", 1);
        BlueCampA.AddUnitByTriggerName("BlueCampA1");
        m_TargetGroups.AddGroup(BlueCampA);
        /// a red group
        CTargetGroup RedCampA = new CTargetGroup(this, CArmies.ARMY_BLUE, "Red Camp A", 1);
        RedCampA.AddUnitByTriggerName("RedCampA1");
        m_TargetGroups.AddGroup(RedCampA);
    }
    /// <summary>
    /// Event: mission has been ended
    /// * The '..Stoped..' is not a spelling mistake here.
    /// * GamePlay.gpBattleStop() is broken, so on Multiplayer Server
    ///   - We call Battle.OnBattleStoped() manually (which calls this mission's and all sub-missions' OnBattleStoped)
    ///   - After that we shut down the whole server with a process.kill()
    ///   see EndMission()
    /// * Entering "battle stop" in a headless server's console will correctly call this.
    /// </summary>
    public override void OnBattleStoped()
    {
        base.OnBattleStoped();
        MissionNumberListener = 0;
        /// cleanup any pending Rearm/Refuel requests
        m_RearmRefuel.CancelAll();
        /// remove all players
        m_MissionPlayers.Clear();
    }
    /// <summary>
    /// On Bomb explosion, check where and if on target increase it's damage (in kg received).
    /// </summary>
    /// <param name=""></param>
    public override void OnBombExplosion(string title, double mass, Point3d pos, AiDamageInitiator initiator, int eventArgInt)
    {
        base.OnBombExplosion(title, mass, pos, initiator, eventArgInt);
        CTargetDamage TargetDamage = null;

        /// Note: Only one single bomb explosion is reported with each call to this function,
        ///       but it could well have damaged / destroyed several targets.
        ///       So mind the if ... else
        ///         * if you have, e.g. a TargetArea within a TargetAirfield
        ///         * BUT it consumes a lot processing resources to check every possible target
        ///           for every single bomb ...

        /// Check if any TargetAirfields were hit (TargetDamage stays null if not)
        if (null != (TargetDamage = m_TargetAirfields.OnBombExplosion(title, mass, pos, initiator, eventArgInt)))
        {
            m_MissionStatus.RegisterTargetDamage(TargetDamage, CArmies.ARMY_ALL, m_Broadcast);
        }
        /// Check if any TargetAreas were hit (TargetDamage stays null if not)
        else if (null != (TargetDamage = m_TargetAreas.OnBombExplosion(title, mass, pos, initiator, eventArgInt)))
        {
            m_MissionStatus.RegisterTargetDamage(TargetDamage, CArmies.ARMY_ALL, m_Broadcast);
        }

        /// Check if any other targets were damaged / destroyed

        /// Example to manually register a damaged/destroyed target with m_MissionStatus
        //TargetDamage = new CTargetDamage();
        //TargetDamage.iArmy = CArmies.ARMY_BLUE;
        //TargetDamage.bDestroyed = false;
        //TargetDamage.iPercent = 50;
        //TargetDamage.iPoints = 1;
        //TargetDamage.sName = "MySpecialTarget";
        //m_MissionStatus.RegisterTargetDamage(TargetDamage, CArmies.ARMY_ALL, m_Broadcast);
    }
    /// <summary>
    /// Communication between the main mission and sub missions (if any)
    /// </summary>
    /// <param name="sMsg"></param>
    /// <param name="Obj"></param>
    /// <returns></returns>
    public override object[] OnIntraMissionsMessage(string sMsg, object[] args = null)
    {
        /// The CServerStats-vX.y.z-Mission.mis 'registers' itself with the main mission:
        if (sMsg.Equals("CServerStats: register"))
        {
            m_ServerStatMis = (args[0] as AMission);
            /// object[0]: string, absolute path to .ini file containing configuration
            m_ServerStatMis.OnIntraMissionsMessage("Start", new object[] { m_sUserDoc + m_sMyFolder + @"modules\ServerStats\CServerStats.ini" });
            /// object[0]: string, path/mission.briefing, relative to CloD-UserDoc
            m_ServerStatMis.OnIntraMissionsMessage("SetMissionFromBriefing", new object[] { m_sMyFolder + Path.GetFileNameWithoutExtension(this.PathMyself) + ".briefing" });
        }
        return null;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="ID"></param>
    /// <param name="menuItemIndex"></param>
    public override void OnOrderMissionMenuSelected(Player CurPlayer, int ID, int menuItemIndex)
    {
        base.OnOrderMissionMenuSelected(CurPlayer, ID, menuItemIndex);
        m_MissionMenu.HandleMenuSelected(CurPlayer, ID, menuItemIndex);
    }
    /// <summary>
    /// Called when
    /// - The Player spawns into the vehicle for the first time ("Create")
    ///   Note: OnPlaceLeave *may* be called before with same parameters (if player stayed connected
    ///   in a battle stop ... battle start sequence).
    /// - Player switches from pilot to another crew position (e.g pilot => bombardier, pilot => gunner)
    /// - Player switches from crew position to another crew position (e.g bombardier => gunner)
    /// It is *not* called
    /// - When player switches back to pilot *not* using "Release Position"
    /// - When player goes back to pilot  using "Release Position"
    /// </summary>
    /// <param name="player"></param>
    /// <param name="actor"></param>
    /// <param name="placeIndex"></param>
    public override void OnPlaceEnter(Player CurPlayer, AiActor actor, int placeIndex)
    {
        base.OnPlaceEnter(CurPlayer, actor, placeIndex);
        /// we do this because a player that had already been connected at OnBattleStart is not caught with OnPlayerConnected
        m_MissionPlayers.OnPlaceEnter(CurPlayer);
        /// Prepare the 4. Mission menu for this player
        m_MissionMenu.CreateMainMenu(CurPlayer);
    }
    /// <summary>
    /// If a player leaves his aircraft, the aircraft *must* be destroyed and
    /// removed from the game to make sure that he (or sombody else) cannot
    /// regain it (prevent "outside views").
    ///
    /// Called when
    /// - When player 'leaves' his vehicle (Briefing page, click on flag, say "yes"; or using "Release Position", default: ALT-F2)
    /// - *AFTER* a player has disconnected (if player had been still in a vehicle when exiting the server)
    /// - Player switches from crew position to another crew position (e.g bombardier => gunner); placeIndex is previous position ('bombardier')
    /// - When player goes back to pilot using "Release Position"; placeIndex is released position (e.g. gunner, bombardier)
    /// - If a player stayed connected in a battle stop ... battle start sequence, and spawns into the vehicle for
    ///   the first time ("Create"); placeIndex is 'pilot'. OnPlaceEnter is called after this with same parameters
    /// It is *not* called
    /// - Player switches from pilot to another crew position (e.g pilot => bombardier, pilot => gunner)
    /// - When player switches back to pilot *not* using "Release Position"
    /// </summary>
    public override void OnPlaceLeave(Player CurPlayer, AiActor actor, int placeIndex)
    {
        base.OnPlaceLeave(CurPlayer, actor, placeIndex);
        m_KillDisusedPlanes.OnPlaceLeave(CurPlayer, actor, placeIndex);
        
        /// cleanup any pending Rearm/Refuel requests of this player
        /// Unfortunately this also hits if a player just changes places within the plane!
        m_RearmRefuel.CancelForPlayer(CurPlayer as IPlayer);
    }
    /// <summary>
    /// Called when a player connects
    /// NOTE: This is *not* invoked if player is already connected when battle starts
    ///       Therefore we double check some stuff in OnPlaceEnter
    /// </summary>
    /// <param name="CurPlayer"></param>
    public override void OnPlayerConnected(Player CurPlayer)
    {
        base.OnPlayerConnected(CurPlayer);
        ///// we do this alsoe on OnPlaceEnter because a player that had already been connected at OnBattleStart is not caught with OnPlayerConnected
        m_MissionPlayers.OnConnect(CurPlayer);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="diagnostic"></param>
    public override void OnPlayerDisconnected(Player CurPlayer, string diagnostic)
    {
        base.OnPlayerDisconnected(CurPlayer, diagnostic);
        m_MissionPlayers.OnDisconnect(CurPlayer, diagnostic);
    }
    /// <summary>
    /// Event: During the mission
    /// </summary>
    public override void OnTickGame()
    {
        base.OnTickGame();
        m_MissionStatus.OnTickGame(Time.tickCounter());
        if (null != m_ServerStatMis)
        {
            /// Update ServerStats (ServerStats takes care of *not* updating 33.35 times per second ;-), see CServerStats.ini)
            m_ServerStatMis.OnIntraMissionsMessage("UpdateMissionTimeLeft", new object[] { m_MissionStatus.GetTimeLeft() });
        }
        /// returns true if mission is already over
        if (m_MissionStatus.IsOver())
        {
            return;
        }
        /// Returns true if mission *just* now has met ending conditions
        /// Note: This returns true only once!
        else if (m_MissionStatus.EndsNow())
        {
            EndMission();
            return;
        }
        /// Broadcast scheduled messages
        m_Broadcast.DoAutoMessages();
    }
    /// <summary>
    /// </summary>
    public override void OnTrigger(int missionNumber, string shortName, bool active)
    {
        base.OnTrigger(missionNumber, shortName, active);
        CTargetDamage TargetDamage = null;

        /// Note: only one single trigger being triggered is reported with each call to
        ///       this function, so no need to check any other triggers once you've found
        ///       the one triggered.
        /// Check if any TargetConvoys were triggered (TargetDamage stays null if not)
        if (null != (TargetDamage = m_TargetConvoys.OnTrigger(missionNumber, shortName, active)))
        {
            m_MissionStatus.RegisterTargetDamage(TargetDamage, CArmies.ARMY_ALL, m_Broadcast);
        }
        /// Check if any Radar headquarters or stations were triggered (TargetDamage stays null if not)
        else if (null != (TargetDamage = m_Radar.OnTrigger(missionNumber, shortName, active)))
        {
            m_MissionStatus.RegisterTargetDamage(TargetDamage, CArmies.ARMY_ALL, m_Broadcast);
        }
        /// Check if any item in a TargetGroup was triggered (TargetDamage stays null if not)
        else if (null != (TargetDamage = m_TargetGroups.OnTrigger(missionNumber, shortName, active)))
        {
            m_MissionStatus.RegisterTargetDamage(TargetDamage, CArmies.ARMY_ALL, m_Broadcast);
        }
        /// Check if any other targets were triggered
        else
        {
            /// Example to manually register a damaged/destroyed target with m_MissionStatus
            //TargetDamage = new CTargetDamage();
            //TargetDamage.iArmy = CArmies.ARMY_RED;
            //TargetDamage.bDestroyed = true;
            //TargetDamage.iPercent = 100;
            //TargetDamage.iPoints = 1;
            //TargetDamage.sName = "MySpecialTarget";
            //m_MissionStatus.RegisterTargetDamage(TargetDamage, CArmies.ARMY_ALL, m_Broadcast);
        }
    }
    ///
    /// ENDING THE BATTLE
    /// ---------------------------------------------------------------------------
    /// <summary>
    /// Shows Result message ("winner is ... " or "ends as a draw ... ")
    /// Initiates game end sequence
    /// 
    /// This function can be called only once!
    /// 
    /// Process is actually killed only after 125 seconds.
    /// 
    /// </summary>
    private void EndMission()
    {
        Timeout(5, () => {
             m_Broadcast.ToChatAndScreen(CArmies.ARMY_ALL, m_MissionStatus.GetMissionResultString());
        });
        Timeout(30, () => {
            m_Broadcast.ToChatAndScreen(CArmies.ARMY_ALL, "Mission is restarting soon!!!");
        });
        Timeout(60, () => {
            m_Broadcast.ToChatAndScreen(CArmies.ARMY_ALL, "Server restarting in 1 minute!!!");
        });
        Timeout(120, () => {
            m_Broadcast.ToChatAndScreen(CArmies.ARMY_ALL, "Mission ended. Please wait 2 minutes to reconnect!!!");
        });
        Timeout(125, () => {
            /// Until we can officially stop a battle with GamePlay.gpBattleStop(), we take this workaround:
            /// - We call Battle.OnBattleStoped() manually (which calls this mission's and all sub-missions' OnBattleStoped)
            ///   We catch all exceptions, so that kill *will* be executed!
            /// - After that we shut down the whole server with a process.kill()
            try
            {
                Battle.OnBattleStoped();
            }
            catch (Exception e)
            {
                LogMessage("Battle.OnBattleStoped, exception: " + e.Message);
            }
            Process.GetCurrentProcess().Kill();
        });
    }
/// 
/// COMPOSING MESSAGES
/// ---------------------------------------------------------------------------
    /// <summary>
    ///
    /// </summary>
    public string GetScoreStringForChat()
    {
        return m_MissionStatus.GetTimeLeftString() + "\n" + m_MissionStatus.GetScoreString(true);
    }
    /// <summary>
    ///
    /// </summary>
    public string GetScoreStringForScreen()
    {
        return m_MissionStatus.GetScoreString(false);
    }
    /// <summary>
    ///
    /// </summary>
    public string GetTimeLeftString()
    {
        return m_MissionStatus.GetTimeLeftString();
    }
    /// <summary>
    ///
    /// </summary>
    public string GetScoreStringBlue()
    {
        return m_MissionStatus.GetScoreString(CArmies.ARMY_BLUE, true);
    }
    /// <summary>
    ///
    /// </summary>
    public string GetScoreStringRed()
    {
        return m_MissionStatus.GetScoreString(CArmies.ARMY_RED, true);
    }
/// 
/// MISCELLANEOUS
/// ---------------------------------------------------------------------------
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sMsg"></param>
    protected void LogMessage(string sMsg)
    {
        GamePlay.gpLogServer(null, "Base-Mission: " + sMsg, new object[] { });
    }
}