///
/// Manage AI Missions
///
/// @requires
///
/// @Author    Team Fusion
/// @Version   1.0.1
///
/// === Example integration into Mission ===
///
/// //$include <put-your-path-here>/CAiMissionCommand-vX.y.z.cs
///
/// ** 1.0.1 **
/// - add: SetMaxPlayersOwnArmy
/// - add: SetMaxConcurrentMissions
/// - add: SetMaxConcurrentMissionsPerArmy
///
///
using System;
using System.Collections;
using System.Collections.Generic;  /// List
using System.Diagnostics;
using maddox.game;                 /// AMission

//$include .\CMissionMenu-v1.0.0.cs
using MissionMenu;

/// <summary>
/// One AI mission
/// </summary> 
public class CAiMission {
    /// <summary>
    /// 
    /// </summary>
    public bool m_bIsRunning = false;
    /// <summary>
    /// 
    /// </summary>
    public int m_iArmy = CArmies.ARMY_NONE;
    /// <summary>
    /// Relative to UserDoc
    /// </summary>
    public string m_sMissionFile;
    /// <summary>
    /// 
    /// </summary>
    public string m_sMenuText;
    /// <summary>
    /// Friend: "Attention: 9 Blenheims, angels 10, start in 10 min at Hawkinge; target: Oye Plage";
    /// Enemy : used Takes sIntel4Enemy
    /// </summary>
    public string[] m_aAlertMsg;
    /// <summary>
    /// Friend: "9 Blenheims, angels 10, from Hawkinge; target: Oye Plage";
    /// Enemy : "Enemy planes detected! Patrol route from AW_22 to AY_21";
    /// </summary>
    public string[] m_aIntelMsg;  
    /// <summary>
    /// 
    /// </summary>
    public Stopwatch m_Time = new Stopwatch();
    /// <summary>
    /// 
    /// </summary>
    /// <param name="iArmy"></param>
    /// <param name="sMissionFile"></param>
    /// <param name="sMenuText">e.g. "9 Blenheims, angels 10, start at Hawkinge; target: Oye Plage"</param>
    /// <param name="sAlert4Friend">"Attention: 9 Blenheims, angels 10, start in 10 min at Hawkinge; target: Oye Plage"</param>
    /// <param name="sIntel4Friend">e.g. "9 Blenheims, angels 10, from Hawkinge; target: Oye Plage"</param>
    /// <param name="sIntel4Enemy">e.g. "Enemy planes detected! Patrol route from AW_22 to AY_21"</param>
    public CAiMission(int iArmy, string sMissionFile, string sMenuText, string sAlert4Friend, string sIntel4Friend, string sIntel4Enemy)
    {
        m_iArmy = iArmy;
        m_sMissionFile = sMissionFile;
        m_sMenuText = sMenuText;
        m_aAlertMsg = new string[CArmies.ARMY_COUNT + 1];
        m_aIntelMsg = new string[CArmies.ARMY_COUNT + 1];
        /// everybody gets the 'enemy' intel ...
        for(int i=0; i <= CArmies.ARMY_COUNT; i++)
        {
            m_aAlertMsg[i] = sIntel4Enemy;
            m_aIntelMsg[i] = sIntel4Enemy;
        }
        /// ... except ourselves
        m_aAlertMsg[m_iArmy] = sAlert4Friend;
        m_aIntelMsg[m_iArmy] = sIntel4Friend;

    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="ParentMission"></param>
    /// <param name="Broadcast"></param>
    public bool Launch(AMission ParentMission, CBroadcast Broadcast)
    {
        try
        {
            ParentMission.GamePlay.gpPostMissionLoad(m_sMissionFile);
        }
        catch(Exception e)
        {
            return false;
        }

        m_bIsRunning = true;
        m_Time.Restart();
        if(null != Broadcast)
        {
            for (int iArmy = 0; iArmy <= CArmies.ARMY_COUNT; iArmy++)
            {
                Broadcast.ToChatAndScreen(iArmy, m_aAlertMsg[iArmy]);
            }
        }
        return true;
    }
    /// <summary>
    /// 
    /// </summary>
    public void OnTimeOut()
    {
        m_bIsRunning = false;
        m_Time.Stop();
    }
}
/// <summary>
/// Manage all AI Missions
/// </summary> 
public class CAiMissionCommand {
    /// <summary>
    /// 
    /// </summary>
    public bool DEBUG_MESSAGES = false;
    /// <summary>
    /// 
    /// </summary>
    protected AMission m_Mission = null;
    /// <summary>
    /// 
    /// </summary>
    protected CBroadcast m_Broadcast = null;
    /// <summary>
    /// 
    /// </summary>
    protected List<CAiMission> m_aMissions = null;
    /// <summary>
    /// 
    /// </summary>
    int m_nMissionInProgress = 0;
    int[] m_aMissionsInProgress = new int[CArmies.ARMY_COUNT + 1] { 0, 0, 0};
    /// <summary>
    /// Maximum number of concurrent missions (all armies)
    /// </summary>
    int m_nCfgMaxMissions = 1;
    /// <summary>
    /// Maximum number of concurrent missions of player's army
    /// </summary>
    int m_nCfgMaxMissionsPerArmy = 0;
    /// <summary>
    /// 
    /// </summary>
    int m_nCfgMaxMissionTime = 3600;
    /// <summary>
    /// 
    /// </summary>
    CMissionPlayers m_MissionPlayers = null;
    /// <summary>
    /// Maximum number of all players (all armies combined) beyond which
    /// no more ai missions are possible
    /// 0 means no limit
    /// </summary>
    int m_nMaxPlayers = 0;
    /// <summary>
    /// Maximum number of players of player's own army beyond which
    /// no more ai missions are possible
    /// 0 means no limit
    /// </summary>
    int m_nMaxPlayersOwnArmy = 0;
    /// <summary>
    /// Constructor
    /// </summary> 
    /// <param name="Mission">The Mission (use 'this')</param>
    public CAiMissionCommand(AMission Mission, CBroadcast Broadcast)
    {
        m_Mission = Mission;
        m_Broadcast = Broadcast;
        m_aMissions = new List<CAiMission>();
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <returns></returns>
    public bool CanStartMission(Player CurPlayer, out string sDenyReason)
    {
        int iPlayerArmy = CurPlayer.Army();
        
        /// test number of missions
        if(0==m_nCfgMaxMissionsPerArmy && m_nCfgMaxMissions <= m_nMissionInProgress)
        {
            sDenyReason = m_nMissionInProgress+" AI mission" + (m_nMissionInProgress > 1 ? "s" : "") + " already in progress. Check Report.";
            return false;
        }
        else if(m_nCfgMaxMissionsPerArmy > 0 && m_nCfgMaxMissionsPerArmy <= m_aMissionsInProgress[iPlayerArmy])
        {
            sDenyReason = m_aMissionsInProgress[iPlayerArmy]+" "+ CArmies.ARMY_COLOURS[iPlayerArmy] +" AI mission" + (m_aMissionsInProgress[iPlayerArmy] > 1 ? "s" : "") + " already in progress. Check Report.";
            return false;
        }
        /// test number of players
        else if (m_nMaxPlayers > 0 && null != m_MissionPlayers && (0==m_nMaxPlayersOwnArmy && m_MissionPlayers.GetConnectedPlayers() >= m_nMaxPlayers))
        {
            sDenyReason = "More than "+ m_nMaxPlayers+" players online.";
            return false;
        }
        else if(m_nMaxPlayersOwnArmy > 0 && null != m_MissionPlayers && m_MissionPlayers.GetConnectedPlayers(iPlayerArmy) >= m_nMaxPlayersOwnArmy)
        {
            sDenyReason = "More than "+ m_nMaxPlayersOwnArmy+" players on your side online.";
            return false;
        }
        sDenyReason = null;
        return true;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <returns></returns>
    public bool CanStartMission(Player CurPlayer)
    {
        string sDenyReason;
        return CanStartMission(CurPlayer, out sDenyReason);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="iArmy"></param>
    /// <param name="sMissionFile"></param>
    /// <param name="sMenuText">e.g. "9 Blenheims, angels 10, start at Hawkinge; target: Oye Plage"</param>
    /// <param name="sAlert4Friend">"Attention: 9 Blenheims, angels 10, start in 10 min at Hawkinge; target: Oye Plage"</param>
    /// <param name="sIntel4Friend">e.g. "9 Blenheims, angels 10, from Hawkinge; target: Oye Plage"</param>
    /// <param name="sIntel4Enemy">e.g. "Enemy planes detected! Patrol route from AW_22 to AY_21"</param>
    public void AddMission(int iArmy, string sMissionFile, string sMenuText, string sAlert4Friend, string sIntel4Friend, string sIntel4Enemy)
    {
        m_aMissions.Add(new CAiMission(iArmy, sMissionFile, sMenuText, sAlert4Friend, sIntel4Friend, sIntel4Enemy));
    }
    /// <summary>
    /// Delegate to populate a CMissionMenu's dynamic Menu.
    /// This is closely tied to OnMissionMenuSelected, as we only have the idx1Item (=the sequence) of the
    /// menu's commands to connect the mission to the selected command.
    /// See CMissionMenu's explanation on Dynamic Menus for technical details.
    /// </summary>
    /// <param name="Menu"></param>
    /// <param name="CurPlayer"></param>
    /// <param name="aExtraData"></param>
    public void OnMissionMenuMakeCommands(MissionMenu.MenuItem Menu, Player CurPlayer, object[] aExtraData)
    {
        bool bEnabled = CanStartMission(CurPlayer);
        /// Not needed (done by CMissionMenu before call): Menu.Clear();
        foreach (CAiMission AiMis in m_aMissions)
        {
            if(CurPlayer.Army() == AiMis.m_iArmy)
            {
                MissionMenu.MenuItem Command = Menu.AddDynamicCommand(AiMis.m_sMenuText);
                Command.m_bEnabled = bEnabled;
            }
        }
    }
    /// <summary>
    /// Delegate to process CMissionMenu's dynamic Menu commad selected.
    /// 
    /// * This is closely tied to OnMissionMenuMakeCommands, as we only have the idx1Item (=the sequence) of the
    ///   menu's commands to connect the mission to the selected command.
    ///   
    /// * ATTENTION: This might also be invoked if the command had been disabled (red color)
    /// 
    /// See CMissionMenu's explanation on Dynamic Menus for technical details.
    ///   
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="idMenu"></param>
    /// <param name="idx1Item"></param>
    /// <returns>false if the menu command is/was actually meant to be disabled</returns>
    public bool OnMissionMenuSelected(Player CurPlayer, int idMenu, int idx1Item)
    {
        string sDenyReason;
        if (!CanStartMission(CurPlayer, out sDenyReason))
        {
            m_Mission.GamePlay.gpHUDLogCenter(new Player[] { CurPlayer }, sDenyReason, null);
            m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, sDenyReason, null);
            return false;
        }
        int idx = 0;
        /// We do exactly the same as in OnMissionMenuMakeCommands,
        /// hoping ;-) that the index will turn out to be same mission.
        foreach (CAiMission AiMis in m_aMissions)
        {
            if (CurPlayer.Army() == AiMis.m_iArmy)
            {
                idx++;
                if(idx == idx1Item)
                {
                    StartMission(CurPlayer, AiMis);
                    break;
                }
            }
        }
        return true;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="aExtraData"></param>
    public void OnMissionMenuGetReport(Player CurPlayer, object[] aExtraData)
    {
        if(0== m_nMissionInProgress)
        {
            string sText = "No AI missions in progress.";
            m_Mission.GamePlay.gpHUDLogCenter(new Player[] { CurPlayer }, sText, null);
            m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, sText, null);
        }
        else
        {
            int iTimeout = 0;
            double dMinTime = m_nCfgMaxMissionTime;
            foreach (CAiMission AiMis in m_aMissions)
            {
                if(AiMis.m_bIsRunning)
                {
                    dMinTime = Math.Min(dMinTime, AiMis.m_Time.Elapsed.TotalSeconds);
                    m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, AiMis.m_aIntelMsg[CurPlayer.Army()], null);
                    m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "Started "+ String.Format("{0:00}", AiMis.m_Time.Elapsed.TotalMinutes)+" minutes ago.", null);
                    m_Mission.Timeout(iTimeout, () => {  m_Mission.GamePlay.gpHUDLogCenter(new Player[] { CurPlayer }, AiMis.m_aIntelMsg[CurPlayer.Army()], null); });
                    iTimeout += 4;
                }
            }
            if(!CanStartMission(CurPlayer))
            {
                string sNext = "A new mission can be started in " + String.Format("{0:00}", Math.Floor((double)(m_nCfgMaxMissionTime - dMinTime) / 60)) + " minutes.";
                m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, sNext, null);
                m_Mission.Timeout(iTimeout, () => { m_Mission.GamePlay.gpHUDLogCenter(new Player[] { CurPlayer }, sNext, null); });
            }
        }
    }
    /// <summary>
    /// Set a maximum of concurrent missions (all armies)
    /// </summary>
    /// <param name="nNumber"></param>
    public void SetMaxConcurrentMissions(int nNumber)
    {
        m_nCfgMaxMissions = nNumber;
    }
    /// <summary>
    /// Set a maximum of concurrent missions per army
    /// This overrides SetMaxConcurrentMissions: if red has 2 missions running and you set this
    /// to 2 then blue will still be allowed to start 2 missions.
    /// </summary>
    /// <param name="nNumber"></param>
    public void SetMaxConcurrentMissionsPerArmy(int nNumber)
    {
        m_nCfgMaxMissionsPerArmy = nNumber;
    }
    /// <summary>
    /// Set a max player (all armies combined) limit beyond which no more missions can be ordered
    /// 0 means no limit
    /// </summary>
    /// <param name="nMaxPlayers"></param>
    /// <param name="MissionPlayers"></param>
    public void SetMaxPlayers(int nMaxPlayers, CMissionPlayers MissionPlayers)
    {
        m_MissionPlayers = MissionPlayers;
        m_nMaxPlayers = nMaxPlayers;
    }
    /// <summary>
    /// Set a max player limit for the player's own army beyond which no more missions can be ordered on that side.
    /// 0 means to ignore this setting
    /// This overrides SetMaxPlayers: if max players is 30 and there are 25 blue players and 10 red players and
    /// you set this to 15 then red will be allowed to start a mission.
    /// </summary>
    /// <param name="nMaxPlayers"></param>
    /// <param name="MissionPlayers"></param>
    public void SetMaxPlayersOwnArmy(int nMaxPlayers, CMissionPlayers MissionPlayers)
    {
        m_MissionPlayers = MissionPlayers;
        m_nMaxPlayersOwnArmy = nMaxPlayers;
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="iInSeconds"></param>
    public void SetMissionTimeout(int iInSeconds)
    {
        m_nCfgMaxMissionTime = iInSeconds;
    }
    /// <summary>
    /// 
    /// </summary>
    protected void StartMission(Player CurPlayer, CAiMission AiMis)
    {
        if(AiMis.Launch(m_Mission, m_Broadcast))
        {
            m_nMissionInProgress++;
            m_aMissionsInProgress[CurPlayer.Army()]++;

            m_Mission.Timeout(m_nCfgMaxMissionTime, () => {
                m_nMissionInProgress--;
                m_aMissionsInProgress[CurPlayer.Army()]--;
                AiMis.OnTimeOut();
                if(CanStartMission(CurPlayer) && null != m_Broadcast)
                {
                    m_Broadcast.ToChatAndScreen(CArmies.ARMY_ALL, "Aircraft are on standby and await mission orders.");
                }
            });
        }
    }
}