///
/// Mission Menu
///
/// Used to return damage info from Target-Classes (e.g. CTargetAirfields).
///
/// @Author    Team Fusion
/// @Version   1.0.0
/// 
/// === Example integration into Mission ===
///
/// //$include <put-your-path-here>/CMissionMenu-vX.y.z.cs
///
using System;
using System.Collections.Generic;

using maddox.game;
//$include ./CArmies-v1.0.1.cs

namespace MissionMenu
{
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="aExtraData"></param>
    public delegate void OnCommandDelegate(Player CurPlayer, object[] aExtraData);
    /// <summary>
    /// 
    /// </summary>
    /// <param name="Menu"></param>
    /// <param name="CurPlayer"></param>
    /// <param name="aExtraData"></param>
    public delegate void OnDynamicMenuDraw(MenuItem Menu, Player CurPlayer, object[] aExtraData);

    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="idMenu"></param>
    /// <param name="idx1Item"></param>
    /// <returns>muts return false if the menu command is/was actually meant to be disabled</returns>
    public delegate bool OnDynamicMenuSelected(Player CurPlayer, int idMenu, int idx1Item);

    /// <summary>
    /// 
    /// </summary>
    public class MenuItem
    {
        /// <summary>
        /// 
        /// </summary>
        public bool m_bEnabled = true;
        /// <summary>
        /// We have a text for every army
        /// </summary>
        public string[] m_aText = new string[CArmies.ARMY_COUNT + 1];
        /// <summary>
        /// 
        /// </summary>
        public List<MenuItem> m_aItems = new List<MenuItem>();
        /// <summary>
        /// 
        /// </summary>
        public bool IsDynamic { get { return m_bIsDynamic; } }
        /// <summary>
        /// 
        /// </summary>
        protected bool m_bIsDynamic = false;
        /// <summary>
        /// 
        /// </summary>
        public OnDynamicMenuDraw m_dgOnDynDraw;
        /// <summary>
        /// 
        /// </summary>
        public OnDynamicMenuSelected m_dgOnDynSelect;
        /// <summary>
        /// 
        /// </summary>
        public bool m_bIsCommand = true;
        /// <summary>
        /// 
        /// </summary>
        public MissionMenu.OnCommandDelegate m_dgOnCmd;
        /// <summary>
        /// 
        /// </summary>
        public object[] m_aExtraData = null;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sText"></param>
        public MenuItem(string sText, bool bIsCommand, object[] aExtraData)
        {
            Init(new string[] { sText }, bIsCommand, aExtraData);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="aText">{TextArmyRed, TextArmyBlue}</param>
        /// <param name="dgOnCommand"></param>
        /// <param name="aExtraData"></param>
        public MenuItem(string[] aText, bool bIsCommand, object[] aExtraData)
        {
            Init(aText, bIsCommand, aExtraData);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="aText"></param>
        /// <returns></returns>
        protected MenuItem AddCommand(string[] aText, object[] aExtraData, bool bIsDynamic)
        {
            if (m_bIsCommand)
            {
                throw new Exception("Menu '" + m_aText[0] + "' is a command. Cannot add sub command.");
            }
            else if(bIsDynamic != m_bIsDynamic)
            {
                if(m_bIsDynamic)
                {
                    throw new Exception("Menu '" + m_aText[0] + "' is dynamic: use AddCommandDynamic");
                }
                else
                {
                    throw new Exception("Menu '" + m_aText[0] + "' is not dynamic: use AddCommand");
                }
            }
            else if (m_aItems.Count >= 9)
            {
                throw new Exception("Menu '" + m_aText[0] + "' already has all 9 possible entries");
            }
            MenuItem Item = new MenuItem(aText, true, aExtraData);
            m_aItems.Add(Item);
            return Item;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        /// <param name="dgOnCommand"></param>
        /// <param name="aExtraData"></param>
        public MenuItem AddCommand(string[] aText, MissionMenu.OnCommandDelegate dgOnCommand, object[] aExtraData = null)
        {
            MenuItem Item = AddCommand(aText, aExtraData, false);
            Item.m_dgOnCmd = dgOnCommand;
            return Item;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sText"></param>
        /// <param name="dgOnCommand"></param>
        /// <param name="aExtraData"></param>
        /// <returns></returns>
        public MenuItem AddCommand(string sText, MissionMenu.OnCommandDelegate dgOnCommand, object[] aExtraData = null)
        {
            return AddCommand(new string[] { sText }, dgOnCommand, aExtraData);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="aText"></param>
        /// <returns></returns>
        public MenuItem AddDynamicCommand(string[] aText)
        {
            MenuItem Item = AddCommand(aText, null, true);
            return Item;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sText"></param>
        /// <returns></returns>
        public MenuItem AddDynamicCommand(string sText)
        {
            MenuItem Item = AddCommand(new string[] { sText }, null, true);
            return Item;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sText"></param>
        /// <returns></returns>
        public MenuItem AddMenu(string sText, object[] aExtraData=null)
        {
            return AddMenu(new string[] { sText }, aExtraData);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="aText">{TextArmyRed, TextArmyBlue}</param>
        /// <param name="aExtraData"></param>
        /// <returns>created MenuItem</returns>
        public MenuItem AddMenu(string[] aText, object[] aExtraData = null)
        {
            if (m_bIsCommand)
            {
                throw new Exception("Menu '" + m_aText[0] + "' is a command. Cannot add sub menu.");
            }
            else if (m_bIsDynamic)
            {
                throw new Exception("Menu '" + m_aText[0] + "' is dynamic: sub menus not supported (yet?)");
            }
            else if (m_aItems.Count >= 9)
            {
                throw new Exception("Menu '" + m_aText[0] + "' already has all 9 possible entries");
            }
            MenuItem Item = new MenuItem(aText, false, aExtraData);

            m_aItems.Add(Item);
            return Item;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="aText"></param>
        /// <param name="dgOnDraw"></param>
        /// <param name="dgOnSelect"></param>
        /// <param name="aExtraData"></param>
        /// <returns></returns>
        public MenuItem AddDynamicMenu(string[] aText, OnDynamicMenuDraw dgOnDraw, OnDynamicMenuSelected dgOnSelect, object[] aExtraData = null)
        {
            MenuItem Item = AddMenu(aText, aExtraData);
            Item.m_bIsDynamic    = true;
            Item.m_dgOnDynDraw   = dgOnDraw;
            Item.m_dgOnDynSelect = dgOnSelect;
            return Item;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sText"></param>
        /// <param name="dgOnDraw"></param>
        /// <param name="dgOnSelect"></param>
        /// <param name="aExtraData"></param>
        /// <returns></returns>
        public MenuItem AddDynamicMenu(string sText, OnDynamicMenuDraw dgOnDraw, OnDynamicMenuSelected dgOnSelect, object[] aExtraData = null)
        {
            return AddDynamicMenu(new string[] { sText }, dgOnDraw, dgOnSelect, aExtraData);
        }
        /// <summary>
        /// Remove all SubItems
        /// </summary>
        public void Clear()
        {
            m_aItems.Clear();
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="aText">{TextArmyRed, TextArmyBlue}</param>
        /// <param name="dgOnCommand"></param>
        /// <param name="aExtraData"></param>
        protected void Init(string[] aText, bool bIsCommand, object[] aExtraData)
        {
            string sText = "Unknown";
            /// aText[0] => m_aText[1] ARMY_RED
            /// aText[1] => m_aText[2] ARMY_BLUE
            for (int i = 0; i < CArmies.ARMY_COUNT; i++)
            {
                if (i < aText.Length)
                {
                    sText = aText[i];
                }
                m_aText[i+1] = sText;
            }
            /// ARMY_NONE
            m_aText[0] = sText;
            m_bIsCommand = bIsCommand;
            m_aExtraData = aExtraData;
        }
    }
    /// <summary>
    /// 
    /// </summary>
    public class CMissionMenu : MissionMenu.MenuItem
    {
        protected AMission m_Mission = null;
        /// <summary>
        /// Constructor
        /// </summary> 
        /// <param name="Mission">The Mission (use 'this')</param>
        public CMissionMenu(AMission Mission) : base("Mission", false, null)
        {
            m_Mission = Mission;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        public void CreateMainMenu(Player CurPlayer)
        {
            if (0 == m_aItems.Count)
            {
                return;
            }
            DrawMenu(CurPlayer, this, 0);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="Menu"></param>
        /// <param name="idMenu"></param>
        protected void DrawMenu(Player CurPlayer, MenuItem Menu, int idMenu)
        {
            if(Menu.IsDynamic)
            {
                Menu.Clear();
                Menu.m_dgOnDynDraw(Menu, CurPlayer, Menu.m_aExtraData);
            }
            if (0 == Menu.m_aItems.Count)
            {
                return;
            }
            /// The direct way to 'draw' Mission menu and sub menus is like this:
            ///     m_Mission.GamePlay.gpSetOrderMissionMenu(
            ///         CurPlayer, 
            ///         bIsSubMenu,                                           // true for anything *not* directly under "4. Mission"
            ///         idMenu,                                               // arbitrary ID to identify the menu created here
            ///         new string[] { "Rearm", "Refuel", "Show damages" },   // array with the texts, here for three menu entries
            ///         new bool[] { true, true, true }                       // array for the menu entries being green(true) or red(false), OnOrderMissionMenuSelected is *always* called
            ///     );
            /// We construct the arrays from our menu tree
            List<string> lstNames = new List<string>();
            List<bool> lstEnabled = new List<bool>();
            foreach (MenuItem Item in Menu.m_aItems)
            {
                lstNames.Add(Item.m_aText[CurPlayer.Army()]);
                lstEnabled.Add(Item.m_bEnabled);
            }
            /// Note on enabled: false just makes the item red, it is called nevertheless - BUT: menu closes too (like idx1Item 255!)
            m_Mission.GamePlay.gpSetOrderMissionMenu(CurPlayer, idMenu != 0, idMenu, lstNames.ToArray(), lstEnabled.ToArray());
            //LogMessage("DrawMenu   ID[" + idMenu + "]");
        }
        /// <summary>
        /// NOTE: Menu/Command enabled: false just makes the item red, it is called nevertheless - BUT: menu closes too (like idx1Item 255!)
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="idMenu"></param>
        /// <param name="idx1Item"></param>
        public void HandleMenuSelected(Player CurPlayer, int idMenu, int idx1Item)
        {
            /// Mission Menu                          ID  iLevel     ID / idx1Item
            /// --------------------------------------------------------------------
            /// 4. Mission                             0
            ///    1. Menu                            10       1      0 / 1
            ///       1. Menu                        110       2     10 / 1
            ///          1. Menu                    1110       3    110 / 1
            ///             1. Command                 -       -   1110 / 1
            ///             2. Command                 -       -   1110 / 2
            ///          2. Menu                    1120       3    110 / 2
            ///             1. Command                 -       -   1120 / 1
            ///          3. Menu                    1130       3    110 / 3
            ///       2. Menu                        120       2     10 / 2
            ///       3. Menu                        130       2     10 / 3
            ///    2. Menu                            20       1      0 / 2
            ///       1. Menu                        210       2     20 / 1
            ///          1. Command                    -       -    210 / 1
            ///          2. Command                    -       -    210 / 2
            ///          3. Menu                    2130       3    210 / 3
            ///             1. Command                 -       -   2130 / 1
            ///    3. Menu                            30       1      0 / 3
            ///       1. Command                       -       -     30 / 1
            ///       2. Command                       -       -     30 / 2

            //LogMessage("HandleMenu A idMenu[" + idMenu + "] idx1Item[" + idx1Item + "]");

            /// The menu addressed by idMenu
            MenuItem Menu = null;
            /// The *parent* of the menu
            MenuItem MenuParent = null;
            int idParent = -1;

            /// Root: Main Menu and no parent
            if(idMenu <= 0)
            {
                Menu = this;
            }
            /// Tree: Determine the menu and it's direct parent
            else
            {
                idParent = (int)Math.Floor((float)idMenu / 10);               /// 21340 => 2134, 2130 => 213, 210 => 21
                idParent = (int)(Math.Floor((float)idParent / 10) * 10);  ///  2134 => 2130,  213 => 210,  21 => 20;

                /// This gives us the level from the ID (1120 => 3)
                int iLevel = 0 == idMenu ? 0 : (int)Math.Floor(Math.Log10(idMenu));
                /// This gives us the ID multiplier based on level (3 => 1000)
                int div   = (int)Math.Pow(10, iLevel);
                int idTmp = idMenu;
                int idx1InParent = -1;

                Menu = this; // start at root
                for (int i = 0; i < iLevel; i++)
                {
                    MenuParent  = Menu;

                    /// Calc the index in parent                               i=0                i=1        i=2       iLevel=3
                    idx1InParent = (int)Math.Floor((double)(idTmp / div)); /// idTmp: 2130 idx2 | 130 idx1 | 30 idx3 | 0
                    idTmp -= idx1InParent * div;                           /// idTmp:  130      |  30      |  0      | 
                    div /= 10;                                             /// div:    100      |  10      |  1      | 


                    //LogMessage("HandleMenu B idx1InParent[" + idx1InParent + "] MenuParent.m_aItems[" + MenuParent.m_aItems.Count+"]");

                    Menu = MenuParent.m_aItems[idx1InParent-1];
                }

                //LogMessage("HandleMenu C iLevel[" + iLevel + "] idParent[" + idParent + "] MenuParent[" + MenuParent.m_aText[0] + "]");
            }
            
            /// 'Command' 0: back up one level, draw the parent
            if (idx1Item == 0)
            {
                if(null==MenuParent)
                {
                    CreateMainMenu(CurPlayer);
                }
                else
                {
                    DrawMenu(CurPlayer, MenuParent, idParent);
                }
            }
            /// 'command' 255: exit menu (tab, escape, or other), go back to 'our' main menu (ID 0)
            else if (idx1Item == 255)
            {
                CreateMainMenu(CurPlayer);
            }
            /// Seems that !enabled commands are just red, but called nevertheless ...
            /// ... but also the whole Tab menu is closed, so we better redraw MainMenu
            else if (!Menu.IsDynamic && !Menu.m_aItems[idx1Item - 1].m_bEnabled)
            {
                CreateMainMenu(CurPlayer);
            }
            else
            {
                MenuItem Item = Menu.m_aItems[idx1Item - 1];
                if (Item.m_bIsCommand)
                {
                    /// With dynamic menus, we cannot store data/callback in the items (remember it's on the server,
                    /// not the client!) so we just give the creator idMenu and idx1Item and hope that he remembers
                    /// what was meant ...
                    if (Menu.IsDynamic)
                    {
                        try
                        {
                            /// Seems that !enabled commands are just red, but called nevertheless ...
                            /// ... but also the whole Tab menu is closed, so we better redraw MainMenu
                            /// As we cannot store even the bEnabled with dynamics we rely on the delegate:
                            /// if the delegate tells us it was meant to be disabled, we redraw Main
                            if (!Menu.m_dgOnDynSelect(CurPlayer, idMenu, idx1Item))
                            {
                                CreateMainMenu(CurPlayer);
                            }
                            else
                            {
                                DrawMenu(CurPlayer, Menu, idMenu);
                            }
                        }
                        catch (Exception e)
                        {
                            LogMessage("ERROR: Menu command '" + Item.m_aText[0] + "' failed on: "+e.Message);
                        }
                    }
                    // static command
                    else
                    {
                        if (null== Item.m_dgOnCmd)
                        {
                            LogMessage("NOTICE: Menu command '" + Item.m_aText[0]+"' has no command callback function defined.");
                        }
                        else try
                        {
                            Item.m_dgOnCmd(CurPlayer, Item.m_aExtraData);
                        }
                        catch (Exception e)
                        {
                            LogMessage("ERROR: Menu command '" + Item.m_aText[0] + "' failed on: " + e.Message);
                        }
                        /// Redraw menu
                        DrawMenu(CurPlayer, Menu, idMenu);
                    }
                }
                /// Item is a sub menu
                else
                {
                    DrawMenu(CurPlayer, Item, idMenu * 10 + idx1Item * 10);
                }
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sMsg"></param>
        protected void LogMessage(string sMsg)
        {
            m_Mission.GamePlay.gpLogServer(null, "CMissionMenu: " + sMsg, new object[] { });
        }
    }
}
