///
/// CMissionPlayers - Keep track of players
/// 
/// @Author  Team Fusion
/// @Version 1.0.1
///
/// 
/// === Example integration into Mission ===
///
/// //$include <put-your-path-here>/CMissionPlayers-vX.y.z.cs
///
/// ** 1.0.1 **
/// - add: GetConnectedPlayers(int iArmy)

using System.Collections.Generic;
using maddox.game;

/// <summary>
/// </summary>
public class CMissionPlayer
{
    public string m_sName = null;
    /// <summary>
    /// We cannot trust a maddox.game.Player Instance to always live.
    /// We may also keep track of players that have disconnected
    /// Or we have a list of players pre battle start to limit access to, restrict sides, etc...
    /// </summary>
    public Player m_mdxPlayer = null;

    public Dictionary<string, object> m_aData;

    public CMissionPlayer(string sName)
    {
        m_sName     = sName;
        m_aData     = new Dictionary<string, object>();
    }
    /// <summary>
    /// All Modules use their classname (e.g. "CRadar") as key (prefix).
    /// </summary>
    /// <param name="sKey"></param>
    /// <param name="Data"></param>
    public void SetData(string sKey, object Data)
    {
        if(m_aData.ContainsKey(sKey))
        {
            m_aData.Remove(sKey);
        }
        m_aData.Add(sKey, Data);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="mdxPlayer"></param>
    public void OnConnect(Player mdxPlayer)
    {
        m_mdxPlayer = mdxPlayer;
    }
    /// <summary>
    /// 
    /// </summary>
    public void OnDisconnect()
    {
        m_mdxPlayer = null;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sKey"></param>
    /// <returns></returns>
    public object GetData(string sKey)
    {
        object Obj = null;
        m_aData.TryGetValue(sKey, out Obj);
        return Obj;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <returns></returns>
    public bool IsDisconnected()
    {
        return (null == m_mdxPlayer);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sKey"></param>
    public void RemoveData(string sKey)
    {
        m_aData.Remove(sKey);
    }
}
/// <summary>
/// Not necessarily only currently *connected* players
/// </summary>
public class CMissionPlayers {
    /// <summary>
    /// 
    /// </summary>
    protected AMission m_Mission;
    /// <summary>
    /// 
    /// </summary>
    public Dictionary<string, CMissionPlayer> m_aPlayers;
    /// <summary>
    /// 
    /// </summary>
    /// <param name="Mission"></param>
    public CMissionPlayers(AMission Mission)
    {
        m_Mission = Mission;
        m_aPlayers = new Dictionary<string, CMissionPlayer>();
    }
    /// <summary>
    /// Returns the number of currently connected players
    /// </summary>
    /// <returns></returns>
    public int GetConnectedPlayers()
    {
        int nConnected = 0;
        foreach(KeyValuePair <string, CMissionPlayer> Item in m_aPlayers)
        {
            if (null != Item.Value.m_mdxPlayer) nConnected++;
        }
        return nConnected;
    }
    /// <summary>
    /// Returns the number of currently connected players of the given army
    /// </summary>
    /// <returns></returns>
    public int GetConnectedPlayers(int iArmy)
    {
        int nConnected = 0;
        foreach(KeyValuePair <string, CMissionPlayer> Item in m_aPlayers)
        {
            if (null != Item.Value.m_mdxPlayer && Item.Value.m_mdxPlayer.Army() == iArmy) nConnected++;
        }
        return nConnected;
    }
    /// <summary>
    /// Adds and (or if already registered, just) sets online status.
    /// </summary>
    /// <param name="mdxPlayer"></param>
    public void OnConnect(Player mdxPlayer)
    {
        CMissionPlayer misPlayer = null;
        if (!m_aPlayers.TryGetValue(mdxPlayer.Name(), out misPlayer))
        {
            misPlayer = new CMissionPlayer(mdxPlayer.Name());
            m_aPlayers.Add(mdxPlayer.Name(), misPlayer);
        }
        misPlayer.OnConnect(mdxPlayer);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="mdxPlayer"></param>
    public void OnPlaceEnter(Player mdxPlayer)
    {
        OnConnect(mdxPlayer);
    }
    /// <summary>
    /// 
    /// </summary>
    public void Clear()
    {
        m_aPlayers.Clear();
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="mdxPlayer"></param>
    /// <param name="diagnostics"></param>
    public void OnDisconnect(Player mdxPlayer, string diagnostics)
    {
        CMissionPlayer misPlayer = null;
        if (m_aPlayers.TryGetValue(mdxPlayer.Name(), out misPlayer))
        {
            misPlayer.OnDisconnect();
        }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="mdxPlayer"></param>
    /// <returns></returns>
    public CMissionPlayer Get(Player mdxPlayer)
    {
        CMissionPlayer misPlayer = null;
        m_aPlayers.TryGetValue(mdxPlayer.Name(), out misPlayer);
        return misPlayer;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sName"></param>
    /// <returns></returns>
    public CMissionPlayer Get(string sName)
    {
        CMissionPlayer misPlayer = null;
        m_aPlayers.TryGetValue(sName, out misPlayer);
        return misPlayer;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="mdxPlayer"></param>
    /// <param name="diagnostic"></param>
    public void Remove(Player mdxPlayer, string diagnostic)
    {
        m_aPlayers.Remove(mdxPlayer.Name());
    }
}
