///
/// CRadar - 
/// 
/// @Author  Team Fusion, Artist
/// @Version 1.0.3
///
/// Originally from NO.401 Wolverine, Kodiak, No.54 REDDOG, ATAG_Freya
/// 
/// TODO:
/// - Landmarks
/// - ??? Determine and use (for status reports only) sectors of observers/headquarters
/// - Broadcast loss of radar coverage area
/// 
/// History
/// ** 1.0.3
/// - chg: replaced deprecated sPathMyself with this.PathMyself
/// ** 1.0.2
/// - fix: a bug with large reports not being displayed in-game
/// ** 1.0.1
/// - fix: bug with negative values in Static### entries
///   
/// === Example integration into Mission ===
///
/// //$include <put-your-path-here>/CRadar-vX.y.z.cs
///
using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;

using maddox.game;
using maddox.game.world;
using maddox.GP;

//$include ./CSectionFile-v1.0.0.cs

public class CRadar
{
    /// <summary>
    /// All Messages
    /// </summary>
    public class CMessages
    {
        /// <summary>
        /// 
        /// </summary>
        protected AMission m_Mission = null;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="Mission"></param>
        public CMessages(AMission Mission)
        {
            m_Mission = Mission;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="aircraft"></param>
        /// <param name="HQ"></param>
        public void ConfirmConnectedToHeadquarter(Player CurPlayer, CRadarHeadquarter HQ)
        {
            AiAircraft ac = (CurPlayer.Place() as AiAircraft);

            ac.SayToGroup(ac.AirGroup(), "This_is");
            m_Mission.Timeout(2, () => {
                ac.SayToGroup(ac.AirGroup(), HQ.m_sCallsign);
            });
            m_Mission.Timeout(4, () => {
                ac.SayToGroup(ac.AirGroup(), "Over");
            });
            m_Mission.Timeout(4, () => {
                m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "This is " + HQ.m_sCallsign + ". Over", null);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="HQ"></param>
        public void ConfirmRequest(Player CurPlayer, CRadarHeadquarter HQ)
        {
            m_Mission.Timeout(2, () =>
            {
                AiAircraft ac = (CurPlayer.Place() as AiAircraft);
                ac.SayToGroup(ac.AirGroup(), "Ground_control_here_Stand_by");
                m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "From {0} to {1}:\nFiltering plotted air groups - stand by...", new object[] { HQ.GetName(), CurPlayer.Name() });
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        public void InformThatHqNotActiveAnymore(Player CurPlayer)
        {
            m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, GetArmyStringFor_HQ(CurPlayer) + " out of service. Select another " + GetArmyStringFor_HQ(CurPlayer) + ".", null);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        public void InformThatPlayerNotConnected(Player CurPlayer)
        {
            m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "Not connected. Please select a "+ GetArmyStringFor_HQ(CurPlayer) + " first.", null);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="HQ"></param>
        public void InformThatPlayerOutOfRange(Player CurPlayer, CRadarHeadquarter HQ)
        {
            m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "Connection to "+ HQ.GetName() + " lost. Select another " + GetArmyStringFor_HQ(CurPlayer) + ".", null);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="HQ"></param>
        public void InformThatHQLostToEnemy(Player CurPlayer, CRadarHeadquarter HQ)
        {
            /// This radio frequency is now broadcasting an enemy signal!
            m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, HQ.GetName(CurPlayer.Army()) + " was lost to enemy. Select another " + GetArmyStringFor_HQ(CurPlayer) + ".", null);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="HQ"></param>
        public void InformThatWaitTimeIsOver(Player CurPlayer, CRadarHeadquarter HQ)
        {
            AiAircraft aircraft = (CurPlayer.Place() as AiAircraft);
            aircraft.SayToGroup(aircraft.AirGroup(), "Tell_when_ready");
            m_Mission.GamePlay.gpLogServer(new[] { CurPlayer }, "From {0} to {1}:\nGround control available.", new object[] { HQ.GetName(), CurPlayer.Name() });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="HQ"></param>
        public void AskToWaitAndRepeatPreviousReport(Player CurPlayer, CReport Report, int nMinInterval)
        {
            m_Mission.Timeout(2, () => {
                m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "Please wait " + String.Format("{0:00}", (int)(nMinInterval/60)) + " minutes between requests.\nLast Report was:\n", null);
                if (0 == Report.m_aLines.Count)
                {
                    m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, GetArmyStringFor_NoEnemyContacts(CurPlayer), null);
                }
                else
                {
                    SendLinesRecursive(CurPlayer, Report, 0);
                }
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="Request"></param>
        public void SendReportOnEnemies(Player CurPlayer, CRadarHeadquarter HQ, CReport Report, int nDelay)
        {
            m_Mission.Timeout(nDelay, () =>
            {
                AiAircraft ac = (CurPlayer.Place() as AiAircraft);
                if(0==Report.m_aLines.Count)
                {
                    m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "From {0} to {1}:\n{2}", new object[] { HQ.GetName(), CurPlayer.Name(), GetArmyStringFor_NoEnemyContacts(CurPlayer) });
                }
                else
                {
                    ac.SayToGroup(ac.AirGroup(), "Attention3");
                    m_Mission.Timeout(2, () =>
                    {
                        ac.SayToGroup(ac.AirGroup(), "Enemy_Position");
                        m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "From {0} to {1}", new object[] { HQ.GetName(), CurPlayer.Name() });
                        SendLinesRecursive(CurPlayer, Report, 0);
                    });
                }
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="Request"></param>
        public void SendReportOnFriendlies(Player CurPlayer, CRadarHeadquarter HQ, CReport Report, int nDelay)
        {
            m_Mission.Timeout(nDelay, () =>
            {
                AiAircraft ac = (CurPlayer.Place() as AiAircraft);
                if (0 == Report.m_aLines.Count)
                {
                    m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "From {0} to {1}:\n{2}", new object[] { HQ.GetName(), CurPlayer.Name(), GetArmyStringFor_NoFriendlyContacts(CurPlayer) });
                }
                else
                {
                    ac.SayToGroup(ac.AirGroup(), "Attention3");
                    m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, "From {0} to {1}:", new object[] { HQ.GetName(), CurPlayer.Name() });
                    SendLinesRecursive(CurPlayer, Report, 0);
                }
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="Pulks"></param>
        /// <param name="CurPlayer"></param>
        /// <returns></returns>
        public CReport CreateReport(CAircraftPulks Pulks, Player CurPlayer, bool bOnFriendly)
        {
            CReport Report = new CReport();
            if (0 == Pulks.m_aPulks.Count)
            {
                return Report;
            }

            /// The key takes the location, heading, vector (but *not* the number of aircraft)
            /// The message  is with number.
            /// This prevents multiple reports on same location/heading
            Dictionary<string, string> aMessages = new Dictionary<string, string>();
            string sContactType = "FRIENDLIES";
            Point2d playerLocation = new Point2d(CurPlayer.Place().Pos().x, CurPlayer.Place().Pos().y);
            double playerspeed = 200;

            if (!bOnFriendly)
            {
                switch (CurPlayer.Army())
                {
                    case CArmies.ARMY_RED:
                        sContactType = "BANDITS";
                        break;
                    case CArmies.ARMY_BLUE:
                        sContactType = "INDIANER";
                        break;
                    default:
                        sContactType = "ENEMIES";
                        break;
                }
            }

            foreach (CAircraftPulk Pulk in Pulks.m_aPulks)
            {
                string sMessage;
                string sSector = m_Mission.GamePlay.gpSectorName(Pulk.m_Airgroup.Pos().x, Pulk.m_Airgroup.Pos().y);
                Vector2d vecAirgroup = new Vector2d(Pulk.m_Airgroup.Vwld().x, Pulk.m_Airgroup.Vwld().y);
                Point2d PosAirgroup = new Point2d(Pulk.m_Airgroup.Pos().x, Pulk.m_Airgroup.Pos().y);

                switch (CurPlayer.Army())
                {
                    case CArmies.ARMY_RED:
                        sMessage = string.Format(
                            "{0} IN SECTOR {1} AT ANGELS {2}, HEADING {3} - VECTOR {4} TO INTERCEPT",
                            sContactType,
                            sSector,
                            CRadarMath.AltitudeToAngels(Pulk.m_Airgroup.Pos().z),
                            CRadarMath.GetDegreesIn10Step(CRadarMath.CalculateBearingDegree(vecAirgroup)),
                            CRadarMath.GetDegreesIn10Step(CRadarMath.CalculateBearingFromOrigin(CRadarMath.CalculateInterceptionPoint(PosAirgroup, vecAirgroup, playerLocation, playerspeed), playerLocation))
                        );
                        break;
                    //case CArmies.ARMY_BLUE:
                    default:
                        sMessage = string.Format(
                            "{0} IN SECTOR {1} AT {2} METERS, HEADING {3} - VECTOR {4} TO INTERCEPT",
                            sContactType,
                            sSector,
                            CRadarMath.AltitudeToMeters(Pulk.m_Airgroup.Pos().z),
                            CRadarMath.GetDegreesIn10Step(CRadarMath.CalculateBearingDegree(vecAirgroup)),
                            CRadarMath.GetDegreesIn10Step(CRadarMath.CalculateBearingFromOrigin(CRadarMath.CalculateInterceptionPoint(PosAirgroup, vecAirgroup, playerLocation, playerspeed), playerLocation))
                        );
                        break;
                }
                if (!aMessages.ContainsKey(sMessage))
                {
                    aMessages.Add(sMessage, string.Format("** {0}+ " + sMessage, RoundAmountOfAircraft(Pulk.m_nAircraft)));
                }
            }

            foreach(KeyValuePair<string, string>Item in aMessages)
            {
                Report.m_aLines.Add(Item.Value);
            }
            return Report;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <returns></returns>
        public string GetArmyStringFor_Bandit(Player CurPlayer)
        {
            switch (CurPlayer.Army())
            {
                case CArmies.ARMY_BLUE:
                    return "Indianer";
                default:
                    return "Bandits";
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <returns></returns>
        public string GetArmyStringFor_HQ(Player CurPlayer)
        {
            switch(CurPlayer.Army())
            {
                case CArmies.ARMY_BLUE:
                    return "Leitstelle";
                default:
                    return "Ground Control HQ";
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <returns></returns>
        public string GetArmyStringFor_NoEnemyContacts(Player CurPlayer)
        {
            switch (CurPlayer.Army())
            {
                case CArmies.ARMY_RED:
                    return "I have no trade for you.";
                default:
                    return "No enemy contacts to report.";
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <returns></returns>
        public string GetArmyStringFor_NoFriendlyContacts(Player CurPlayer)
        {
            switch (CurPlayer.Army())
            {
                case CArmies.ARMY_RED:
                    return "I have no trade for you.";
                default:
                    return "No friendly contacts to report.";
            }
        }
        /// <summary>
        /// Rounds to 1, 3, 5, 10, 20, 30, ....
        /// </summary>
        /// <param name="number"></param>
        /// <returns></returns>
        protected int RoundAmountOfAircraft(int number)
        {
            int firstDecimal = 0;
            int higherDecimal = 0;

            higherDecimal = Math.DivRem(number, 10, out firstDecimal);

            /// more than 10: 11-13, 21-23, 31-33, ...
            if (firstDecimal >= 1 && firstDecimal <= 3 && higherDecimal > 0)
            {
                firstDecimal = 0; /// => 10, 20, 30, ...
            }
            /// 3, 4, 14, 24, 34, ...
            else if (firstDecimal > 2 && firstDecimal <= 4)
            {
                firstDecimal = 3; /// => 3, 10, 20, 30, ...
            }
            /// 5, 6, 7, 15, 16, 17, 25, 26, 27, ...
            else if (firstDecimal > 4 && firstDecimal <= 7)
            {
                firstDecimal = 5; /// => 5, 10, 20, 30, ...
            }
            /// 7, 8, 9
            else if (firstDecimal > 7)
            {
                higherDecimal += 1; /// 10, 20, 30, ...
            }

            if (higherDecimal > 0)
            {
                return (int)higherDecimal * 10;
            }
            else
            {
                return (int)firstDecimal;
            }
        }
        /// <summary>
        /// We display report line by line.
        /// Also, there's a limit on the length of single a message.
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <param name="Report"></param>
        /// <param name="idx"></param>
        protected void SendLinesRecursive(Player CurPlayer, CReport Report, int idx)
        {
            m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, Report.m_aLines[idx], null);
            if (++idx < Report.m_aLines.Count)
            {
                m_Mission.Timeout(0.7, () =>
                {
                    SendLinesRecursive(CurPlayer, Report, idx);
                });
            }
        }
    }
    /// <summary>
    /// A Radar Report.
    /// </summary>
    public class CReport
    {
        /// <summary>
        /// 
        /// </summary>
        public List<string> m_aLines = new List<string>();
    }
    /// <summary>
    /// Calculations and Conversions
    /// </summary>
    public static class CRadarMath
    {
        /// <summary>
        /// 
        /// </summary>
        /// <param name="degrees"></param>
        /// <returns></returns>
        public static int GetDegreesIn10Step(double degrees)
        {
            degrees = Math.Round((degrees / 10), MidpointRounding.AwayFromZero) * 10;

            if ((int)degrees == 360)
            {
                degrees = 0.0;
            }
            return (int)degrees;
        }
        /// <summary>
        /// to get the correct bearing its nessesary to make a litte enter the matrix operation.
        /// the Vector2d.direction() (same as atan2) has 0° at the x-axis and goes counter clockwise, but we need 0° at the y-axis
        /// and clockwise direction
        /// so to convert it we need 
        /// |0 1| |x|   |0*x + 1*y|    |y|
        /// |   | | | = |         | =  | |   // ok not very surprising ;)
        /// |1 0| |y|   |1*x + 0*y|    |x|
        /// </summary>
        /// <param name="vector"></param>
        /// <returns></returns>
        public static double CalculateBearingDegree(Vector3d vector)
        {
            Vector2d matVector = new Vector2d(vector.y, vector.x);
            /// the value of direction is in rad so we need *180/Pi to get the value in degrees
            return matVector.direction() * 180.0 / Math.PI;
        }
        /// <summary>
        /// to get the correct bearing its nessesary to make a litte enter the matrix operation.
        /// the Vector2d.direction() (same as atan2) has 0° at the x-axis and goes counter clockwise, but we need 0° at the y-axis
        /// and clockwise direction
        /// so to convert it we need 
        /// |0 1| |x|   |0*x + 1*y|    |y|
        /// |   | | | = |         | =  | |   // ok not very surprising ;)
        /// |1 0| |y|   |1*x + 0*y|    |x|
        /// </summary>
        /// <param name="vector"></param>
        /// <returns></returns>
        public static double CalculateBearingDegree(Vector2d vector)
        {
            Vector2d newVector = new Vector2d(vector.y, vector.x);
            /// the value of direction is in rad so we need *180/Pi to get the value in degrees
            return newVector.direction() * 180.0 / Math.PI;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="targetLocation"></param>
        /// <param name="originLocation"></param>
        /// <returns></returns>
        public static double CalculateBearingFromOrigin(Point2d targetLocation, Point2d originLocation)
        {
            double deltaX = targetLocation.x - originLocation.x;
            double deltaY = targetLocation.y - originLocation.y;

            double bearing = Math.Atan2(deltaX, deltaY);
            bearing = bearing * (180.0 / Math.PI);

            return (bearing > 0.0 ? bearing : (360.0 + bearing));
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="targetLocation"></param>
        /// <param name="originLocation"></param>
        /// <returns></returns>
        public static double CalculateBearingFromOrigin(Point3d targetLocation, Point3d originLocation)
        {
            double deltaX = targetLocation.x - originLocation.x;
            double deltaY = targetLocation.y - originLocation.y;

            double bearing = Math.Atan2(deltaX, deltaY);
            bearing = bearing * (180.0 / Math.PI);

            return (bearing > 0.0 ? bearing : (360.0 + bearing));
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="a"></param>
        /// <param name="v"></param>
        /// <param name="b"></param>
        /// <param name="s"></param>
        /// <returns></returns>
        public static Point2d CalculateInterceptionPoint(Point2d a, Vector2d v, Point2d b, double s)
        {
            /// a = point2d position of object to be interecepted - RDF variable airgroupPos found in checkSectors
            /// v = vector2d of the object to be intercepted - RDF Variable airgroupVector found in checkSectors
            /// b = point2d position of the interceptor - player.place(0).pos() ??
            /// s = speed of the interceptor - return speed from vector2d of player?
            Point2d pdefault = new Point2d(0, 0);
            double ox = a.x - b.x;
            double oy = a.y - b.y;

            double h1 = v.x * v.x + v.y * v.y - s * s;
            double h2 = ox * v.x + oy * v.y;
            double t;
            if (h1 == 0)
            {
                /// problem collapses into a simple linear equation 
                t = -(ox * ox + oy * oy) / 2 * h2;
            }
            else
            {
                /// solve the quadratic equation
                double minusPHalf = -h2 / h1;

                double discriminant = minusPHalf * minusPHalf - (ox * ox + oy * oy) / h1; // term in brackets is h3
                if (discriminant < 0)
                {
                    /// no (real) solution then...
                    return pdefault;
                }

                double root = Math.Sqrt(discriminant);

                double t1 = minusPHalf + root;
                double t2 = minusPHalf - root;

                double tMin = Math.Min(t1, t2);
                double tMax = Math.Max(t1, t2);

                t = tMin > 0 ? tMin : tMax; // get the smaller of the two times, unless it's negative
                if (t < 0)
                {
                    /// we don't want a solution in the past
                    return pdefault;
                }
            }

            /// calculate the point of interception using the found intercept time and return it
            return new Point2d(a.x + t * v.x, a.y + t * v.y);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="altitude"></param>
        /// <returns></returns>
        public static int AltitudeToAngels(double altitude)
        {
            double altAngels = (altitude / 0.3048) / 1000;

            if (altAngels > 1)
            {
                altAngels = Math.Round(altAngels, MidpointRounding.AwayFromZero);
            }
            else
            {
                altAngels = 1;
            }

            return (int)altAngels;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="altitude"></param>
        /// <returns></returns>
        public static int AltitudeToMeters(double altitude)
        {
            double altMeters = altitude;

            if (altMeters > 500)
            {
                altMeters = Math.Round(altMeters / 100.0, MidpointRounding.AwayFromZero) * 100;
            }
            else
            {
                altMeters = 500;
            }

            return (int)altMeters;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="degrees"></param>
        /// <returns></returns>
        public static string DegreesToWindRose(double degrees)
        {
            String[] directions = { "North", "North East", "East", "South East", "South", "South West", "West", "North West", "North" };
            return directions[(int)Math.Round((((double)degrees % 360) / 45))];
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="distance"></param>
        /// <returns></returns>
        public static int MeterToMiles(double distance)
        {
            double distanceMiles = 0;
            distanceMiles = Math.Round(((distance / 1609.3426)), 0, MidpointRounding.AwayFromZero);   // distance in Miles

            return (int)distanceMiles;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="distance"></param>
        /// <returns></returns>
        public static int MeterToKilometers(double distance)
        {
            double distanceKilometers = 0;
            distanceKilometers = Math.Round(((distance / 1000)), 0, MidpointRounding.AwayFromZero);   // distance in Kilometers

            return (int)distanceKilometers;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="dMeter"></param>
        /// <param name="iArmy"></param>
        /// <returns></returns>
        public static string MetersToArmy(double dMeter, int iArmy)
        {
            switch(iArmy)
            {
                case CArmies.ARMY_RED:
                    return MeterToMiles(dMeter).ToString() + " km";
                default:
                    return MeterToKilometers(dMeter).ToString() + " miles";
            }
        }
    }
    /// <summary>
    /// Aircraft 'Pulk'
    /// </summary>
    public class CAircraftPulk
    {
        public AiAirGroup m_Airgroup = null;
        public int m_nAircraft = 0;
        public CAircraftPulk(AiAirGroup Airgroup, int nAircraft)
        {
            m_Airgroup = Airgroup;
            m_nAircraft = nAircraft;
        }
    }
    /// <summary>
    /// Aircraft 'Pulks'
    /// </summary>
    public class CAircraftPulks
    {
        /// <summary>
        /// 
        /// </summary>
        public bool DEBUG_MESSAGES = false;
        /// <summary>
        /// 
        /// </summary>
        protected AMission m_Mission = null;
        /// <summary>
        /// 
        /// </summary>
        public List<CAircraftPulk> m_aPulks;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="Mission"></param>
        public CAircraftPulks(AMission Mission)
        {
            m_Mission = Mission;
            m_aPulks = new List<CAircraftPulk>();
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="HQ"></param>
        /// <param name="nMinAircraftInPulk"></param>
        /// <param name="dMinObservableHeight"></param>
        public void Filter(CRadarHeadquarter HQ, int nMinAircraftInPulk, double dMinObservableHeight)
        {
            if (DEBUG_MESSAGES) LogMessage("Filter pulks: start");
            for (int idx = m_aPulks.Count - 1; idx >= 0; idx--)
            {
                if (DEBUG_MESSAGES) LogMessage("Filter pulks, pulk #"+idx);
                CAircraftPulk Item = m_aPulks[idx];

                if (Item.m_nAircraft < nMinAircraftInPulk)
                {
                    m_aPulks.RemoveAt(idx);
                    if (DEBUG_MESSAGES) LogMessage("Filtered pulk #"+idx+": too few ac");
                    continue;
                }
                if (Item.m_Airgroup.Pos().z < dMinObservableHeight)
                {
                    m_aPulks.RemoveAt(idx);
                    if (DEBUG_MESSAGES) LogMessage("Filtered pulk #" + idx + ": too low");
                    continue;
                }
                if (!HQ.IsObservable(Item.m_Airgroup.Pos().x, Item.m_Airgroup.Pos().y))
                {
                    m_aPulks.RemoveAt(idx);
                    if (DEBUG_MESSAGES) LogMessage("Filtered pulk #" + idx + ": out of range");
                    continue;
                }
                if (DEBUG_MESSAGES) LogMessage("Kept pulk #" + idx);
            }
            if (DEBUG_MESSAGES) LogMessage("Filtered pulks: " + m_aPulks.Count + " left");
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        public void LoadEnemies(Player CurPlayer, double dPulkMaxRadius)
        {
            for (int iArmy = 1; iArmy <= CArmies.ARMY_COUNT; iArmy++)
            {
                if (CurPlayer.Army() != iArmy)
                {
                    AddAircraftPulks(m_Mission.GamePlay.gpAirGroups(iArmy), dPulkMaxRadius);
                }
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        public void LoadFriendlies(Player CurPlayer, double dPulkMaxRadius)
        {
            AddAircraftPulks(m_Mission.GamePlay.gpAirGroups(CurPlayer.Army()), dPulkMaxRadius);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="Airgroups"></param>
        protected void AddAircraftPulks(AiAirGroup[] Airgroups, double dPulkMaxRadius)
        {

            if (DEBUG_MESSAGES) LogMessage("gather pulks, start " + Airgroups.Length + " airgroups");

            if (null != Airgroups  && 0 < Airgroups.Length)
            {
                Dictionary<AiAirGroup, List<AiAirGroup>> aTmpPulks = new Dictionary<AiAirGroup, List<AiAirGroup>>();

                for (int idx = 0; idx < Airgroups.Length; idx++)
                {
                    if (DEBUG_MESSAGES) LogMessage("gather pulks, airgroup #"+idx);
                    bool bAddedToPulk = false;
                    foreach (var PulkItem in aTmpPulks)
                    {
                        Point3d PosAirgroup = PulkItem.Key.Pos();

                        if (Airgroups[idx].Pos().distance(ref PosAirgroup) < dPulkMaxRadius)
                        {
                            PulkItem.Value.Add(Airgroups[idx]);
                            bAddedToPulk = true;
                        }
                    }
                    if (!bAddedToPulk)
                    {
                        aTmpPulks.Add(Airgroups[idx], new List<AiAirGroup>());
                    }
                }

                foreach (var PulkItem in aTmpPulks)
                {
                    int nPlanes = PulkItem.Key.NOfAirc;
                    if (PulkItem.Value.Count > 0)
                    {
                        PulkItem.Value.ForEach(item => {
                            nPlanes += item.NOfAirc;
                        });
                    }
                    if (DEBUG_MESSAGES) LogMessage("gather pulks, add pulk with " + nPlanes + " aircraft");
                    m_aPulks.Add(new CAircraftPulk(PulkItem.Key, nPlanes));
                }
            }
            if (DEBUG_MESSAGES) LogMessage("gather pulks done: " + m_aPulks.Count+" pulks");
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sMsg"></param>
        protected void LogMessage(string sMsg)
        {
            m_Mission.GamePlay.gpLogServer(null, this.GetType().Name + ": " + sMsg, new object[] { });
        }
    }
    /// <summary>
    /// Stored with CMissionPlayer.SetData to remember last request
    /// </summary>
    public class CPlayerRequest
    {
        public DateTime m_Time;
        public CReport m_Report = null;
    };
    /// <summary>
    /// 
    /// </summary>
    public class CMapInfo {
        /// <summary>
        /// 
        /// </summary>
        public double m_dSectorSize = 0;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sfMisFile"></param>
        public void Read(CSectionFile sfMisFile)
        {
            ReadBattleArea(sfMisFile);
        }
        protected void ReadBattleArea(CSectionFile sfMisFile)
        {
            string sEntry = sfMisFile.Get("MAIN", "BattleArea", null);
            if (null == sEntry)
            {
                throw new Exception("not BattleArea entry");
            }
            /// BattleArea 8500 8500 370000 370000 10000
            ///                        1:left      2:bottom    3:width     4:height    5:sector size
            Regex rx = new Regex(@"^\s*([0-9.]+)\s+([0-9.]+)\s+([0-9.]+)\s+([0-9.]+)\s+([0-9.]+)");
            MatchCollection aMatches = rx.Matches(sEntry);
            if (1 != aMatches.Count)
            {
                throw new Exception(this.GetType().Name + ": 'BattleArea' entry not decipherable");
            }
            m_dSectorSize = double.Parse(aMatches[0].Groups[5].Value);
        }
    }
    /// <summary>
    /// Base class for Observer and Headquarter
    /// </summary>
    public class CRadarObject
    {
        /// <summary>
        /// 
        /// </summary>
        protected AMission m_Mission = null;
        /// <summary>
        /// The static object representing the observer/headquarter ("static99")
        /// </summary>
        public string m_sStaticKey = null;
        /// <summary>
        /// is this destroyable
        /// </summary>
        public bool m_bDestroyable = true;
        /// <summary>
        /// 
        /// </summary>
        public Point2d m_LocationCoords;
        /// <summary>
        /// 
        /// </summary>
        public string m_sTrigger = null;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="Mission"></param>
        /// <param name="sStaticKey"></param>
        /// <param name="bDestroyable"></param>
        /// <param name="nPoints"></param>
        public CRadarObject(AMission Mission, string sStaticKey, bool bDestroyable)
        {
            m_Mission = Mission;
            m_sStaticKey = sStaticKey;
            m_bDestroyable = bDestroyable;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sfMisFile"></param>
        /// <param name="aTrigger"></param>
        public virtual void Activate(CSectionFile sfMisFile, List<CRadarObject> aTrigger)
        {
            SetLocation(sfMisFile);
            if (m_bDestroyable)
            {
                aTrigger.Add(this);
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public string GetSector()
        {
            return m_Mission.GamePlay.gpSectorName(m_LocationCoords.x, m_LocationCoords.y);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sMsg"></param>
        protected void LogMessage(string sMsg)
        {
            m_Mission.GamePlay.gpLogServer(null, this.GetType().Name+": " + sMsg, new object[] { });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sfMisFile"></param>
        protected void SetLocation(CSectionFile sfMisFile)
        {
            string sEntry = sfMisFile.Get("Stationary", m_sStaticKey, null);
            if (null == sEntry)
            {
                string sErr = this.GetType().Name + " '" + m_sStaticKey + "' not found in .mis file, section [Stationary]";
                LogMessage("ERROR: " + sErr);
                throw new Exception(sErr);
            }
            ///                                   xpos      ypos      rotate
            /// Stationary.Radar.EnglishRadar1 gb 302103.00 113807.81 0.00
            ///                        1:type    2:army    3:xpos    4:ypos    
            Regex rx = new Regex(@"^\s*([^ ]+)\s+([^ ]+)\s+([^ ]+)\s+([^ ]+)\s+");
            MatchCollection aMatches = rx.Matches(sEntry);
            if (1 != aMatches.Count)
            {
                string sErr = this.GetType().Name + " '" + m_sStaticKey + "' entry not decipherable for position";
                LogMessage("ERROR: " + sErr);
                throw new Exception(sErr);
            }
            m_LocationCoords.set(
                Double.Parse(aMatches[0].Groups[3].Value),
                Double.Parse(aMatches[0].Groups[4].Value)
            );
        }
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "id["+ m_sStaticKey + "] idTrg["+ (m_sTrigger ?? "none") +"] pos[" + m_LocationCoords.ToString() + "]";
        }
    }
    /// <summary>
    /// Observer, Radar Station
    /// </summary>
    public class CRadarObserver : CRadarObject
    {
        /// <summary>
        /// radius of area covered around the observer station
        /// </summary>
        public double m_dRadius = 0.0f;
        /// <summary>
        /// 
        /// </summary>
        public string m_sName = null;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="Mission"></param>
        /// <param name="sStaticKey"></param>
        /// <param name="bDestroyable"></param>
        /// <param name="nPoints"></param>
        /// <param name="sName"></param>
        /// <param name="dRadius"></param>
        public CRadarObserver(AMission Mission, string sStaticKey, string sName, double dRadius, bool bDestroyable) : base(Mission, sStaticKey, bDestroyable)
        {
            m_sName = sName;
            m_dRadius = dRadius;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public bool IsObservable(double x, double y)
        {
            Point2d Pos = new Point2d(x, y);
            if (m_LocationCoords.distance(ref Pos) <= m_dRadius)
            {
                return true;
            }
            return false;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "Observer '" + m_sName + "' Radius[" + m_dRadius + "] "+base.ToString();
        }
    }
    /// <summary>
    /// Headquarter: Collection of Observers
    /// </summary>
    public class CRadarHeadquarter : CRadarObject
    {
        /// <summary>
        /// if army is not set, it is derived from frontline
        /// Means: It could change over time!
        /// </summary>
        public int m_iArmy = CArmies.ARMY_NONE;
        /// <summary>
        /// 
        /// </summary>
        public string m_sNameIfBlue = null;
        /// <summary>
        /// 
        /// </summary>
        public string m_sNameIfRed = null;
        /// <summary>
        /// 
        /// </summary>
        public string m_sCallsign = null;
        /// <summary>
        /// 
        /// </summary>
        public double m_dRange = 0.0f;
        /// <summary>
        /// 
        /// </summary>
        public int m_nPoints = 0;
        /// <summary>
        /// 
        /// </summary>
        public List<CRadarObserver> m_aObservers;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sStaticKey"></param>
        /// <param name="bDestroyable"></param>
        /// <param name="iArmy"></param>
        /// <param name="sNameIfBlue"></param>
        /// <param name="sNameIfRed"></param>
        /// <param name="sCallsign"></param>
        /// <param name="dRange"></param>
        public CRadarHeadquarter(AMission Mission, string sStaticKey, int iArmy, string sNameIfBlue, string sNameIfRed, string sCallsign, double dRange, int nPoints, bool bDestroyable) : base(Mission, sStaticKey, bDestroyable)
        {
            m_iArmy = iArmy;
            m_sNameIfBlue = sNameIfBlue;
            m_sNameIfRed = sNameIfRed;
            m_sCallsign = sCallsign;
            m_dRange = dRange;
            m_nPoints = nPoints;

            m_aObservers = new List<CRadarObserver>();
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sStaticKey"></param>
        /// <param name="bDestroyable"></param>
        /// <param name="nPoints"></param>
        /// <param name="sName"></param>
        /// <param name="dRadius"></param>
        public void AddObserver(string sStaticKey, string sName, double dRadius, bool bDestroyable)
        {
            m_aObservers.Add(new CRadarObserver(m_Mission, sStaticKey, sName, dRadius, bDestroyable));
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="Mission"></param>
        /// <param name="sfMisFile"></param>
        /// <param name="aTrigger"></param>
        public override void Activate(CSectionFile sfMisFile, List<CRadarObject> aTrigger)
        {
            base.Activate(sfMisFile, aTrigger);
            foreach (CRadarObserver Observer in m_aObservers)
            {
                Observer.Activate(sfMisFile, aTrigger);
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public string GetName(int iArmy = CArmies.ARMY_NONE)
        {
            if (iArmy == CArmies.ARMY_NONE)
            {
                iArmy = GetOwningArmy();
            }
            switch (iArmy)
            {
                case CArmies.ARMY_BLUE:
                    return m_sNameIfBlue;
                case CArmies.ARMY_RED:
                default:
                    return m_sNameIfRed;
            }
        }
        /// <summary>
        /// Current Owner
        /// </summary>
        /// <returns>Army</returns>
        public int GetOwningArmy()
        {
            if (m_iArmy != CArmies.ARMY_NONE)
            {
                return m_iArmy;
            }
            return m_Mission.GamePlay.gpFrontArmy(m_LocationCoords.x, m_LocationCoords.y);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="CurPlayer"></param>
        /// <returns></returns>
        public bool IsInRange(Player CurPlayer)
        {
            if (null == CurPlayer.Place())
            {
                return false;
            }
            Point2d PosPlayer = new Point2d(CurPlayer.Place().Pos().x, CurPlayer.Place().Pos().y);
            if(m_LocationCoords.distance(ref PosPlayer) > m_dRange)
            {
                return false;
            }
            return true;
        }
        /// <summary>
        /// returns true if the given point would be observable by this station.
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        public bool IsObservable(double x, double y)
        {
            foreach (CRadarObserver Observer in m_aObservers)
            {
                if(Observer.IsObservable(x, y))
                {
                    return true;
                }
            }
            return false;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return "Headquarter '" + m_sNameIfBlue + "'/'"+ m_sNameIfRed + "' dRange[" + m_dRange + "] " + base.ToString();
        }
    }

    /// <summary>
    /// Set to true to get informative log output
    /// </summary>
    public bool DEBUG_MESSAGES = false;
    /// <summary>
    /// 
    /// </summary>
    protected const string TRIGGER_HQ = "CRadarHQ_";
    /// <summary>
    /// 
    /// </summary>
    protected const string TRIGGER_OS = "CRadarOS_";
    /// <summary>
    /// 
    /// </summary>
    protected const string PLAYER_DATA_HQ = "CRadar_HQ";
    /// <summary>
    /// 
    /// </summary>
    protected const string PLAYER_DATA_LASTREQUEST = "CRadar_LastRequest";
    /// <summary>
    /// 
    /// </summary>
    protected AMission m_Mission = null;
    /// <summary>
    /// 
    /// </summary>
    protected CMissionPlayers m_MissionPlayers = null;
    /// <summary>
    /// 
    /// </summary>
    protected CMessages m_Messages = null;
    /// <summary>
    /// 
    /// </summary>
    protected List<CRadarHeadquarter> m_aHeadQuarters;
    /// <summary>
    /// 
    /// </summary>
    protected CMapInfo m_MapInfo = new CMapInfo();
    /// <summary>
    /// 
    /// </summary>
    protected int m_nMinRequestInterval = 120;
    /// <summary>
    /// 
    /// </summary>
    protected int m_nResponseTime = 5;
    /// <summary>
    /// 
    /// </summary>
    protected double m_dPulkMaxRadius = 1400.0f;
    /// <summary>
    /// 
    /// </summary>
    protected int m_nMinAircraftInPulk = 2;
    /// <summary>
    /// 
    /// </summary>
    protected double m_dMinObservableHeight = 400.0f;
    /// <summary>
    /// 
    /// </summary>
    /// <param name="Mission"></param>
    public CRadar(AMission Mission, CMissionPlayers MissionPlayers)
    {
        m_Mission = Mission;
        m_MissionPlayers = MissionPlayers;
        m_aHeadQuarters = new List<CRadarHeadquarter>();
        m_Messages = new CMessages(m_Mission);
    }
    /// <summary>
    /// 
    /// </summary>
    public void Activate()
    {
        List<CRadarObject> aTrigger = new List<CRadarObject>();
        ISectionFile sfTrigger = m_Mission.GamePlay.gpCreateSectionFile();
        int idxHq = 0, idxOs = 0;
        string sUserDoc = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + @"\1C SoftClub\il-2 sturmovik cliffs of dover\";
        string sMyFolder = Path.GetDirectoryName(m_Mission.PathMyself) + @"\";
        string sMissionFile = sUserDoc + sMyFolder + Path.GetFileNameWithoutExtension(m_Mission.PathMyself) + ".mis";
        CSectionFile sfMisFile = new CSectionFile(m_Mission);

        try
        {
            sfMisFile.Load(sMissionFile);
        }
        catch(Exception e)
        {
            LogMessage("ERROR failed to read '" + sMissionFile + "': "+e.Message);
            return;
        }

        m_MapInfo.Read(sfMisFile);

        /// Check, Initialize, and activate all Headquarters and their stations
        /// Fills aTrigger with (destroyable) RadarObjects to create triggers for
        foreach (CRadarHeadquarter HQ in m_aHeadQuarters)
        {
            HQ.Activate(sfMisFile, aTrigger);
        }

        /// Create the (SectionFile format) trigger ...
        foreach(CRadarObject Obj in aTrigger)
        {
            if(Obj.GetType() == typeof(CRadarHeadquarter))
            {
                Obj.m_sTrigger = TRIGGER_HQ + string.Format("{0:00}", idxHq++);
            }
            else if (Obj.GetType() == typeof(CRadarObserver))
            {
                Obj.m_sTrigger = TRIGGER_OS + string.Format("{0:00}", idxOs++);
            }
            Obj.m_sTrigger += "_Destroyed";

            /// TGroundDestroyedTrigger 50% destroyed in radius 100m
            sfTrigger.add("Trigger",
                Obj.m_sTrigger,
                " TGroundDestroyed 50 " + Obj.m_LocationCoords.x.ToString() + " " + Obj.m_LocationCoords.y.ToString() + " 100"
            );
        }
        /// ... and load triggers into mission
        m_Mission.GamePlay.gpPostMissionLoad(sfTrigger);


        if (DEBUG_MESSAGES)
        {
            LogMessage("\n=== RADAR ===");
            foreach (CRadarHeadquarter HQ in m_aHeadQuarters)
            {
                LogMessage(HQ.ToString());
                foreach (CRadarObserver OS in HQ.m_aObservers)
                {
                    LogMessage(" - "+OS.ToString());
                }
            }
        }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sIdStationary">The static object representing the Headquarter (.mis file, [stationary])</param>
    /// <param name="iArmy">if you use ARMY_NONE, owning army is determined by frontline/position</param>
    /// <param name="sNameIfBlue">Name if blue (fronts may shift)</param>
    /// <param name="sNameIfRed">Name if red  (fronts may shift)</param>
    /// <param name="sCallsign">Callsign (also audio file)</param>
    /// <param name="dRange"></param>
    /// <returns>Headquarte object</returns>
    public CRadarHeadquarter AddHeadquarter(string sStaticKey, int iArmy, string sNameIfBlue, string sNameIfRed, string sCallsign, double dRange, int nPoints, bool bDestroyable)
    {
        CRadarHeadquarter HQ = new CRadarHeadquarter(m_Mission, sStaticKey, iArmy, sNameIfBlue, sNameIfRed, sCallsign, dRange, nPoints, bDestroyable);
        m_aHeadQuarters.Add(HQ);
        return HQ;
    }
    /// <summary>
    /// Delegate to populate a CMissionMenu's dynamic Menu.
    /// * This is closely tied to OnMissionMenuSelectHQ_Select, as we only have the idx1Item (=the sequence) of the
    ///   menu's commands to connect to the selected command.
    ///   
    /// See CMissionMenu's explanation on Dynamic Menus for technical details.
    /// </summary>
    /// <param name="Menu"></param>
    /// <param name="CurPlayer"></param>
    /// <param name="aExtraData"></param>
    public void OnMissionMenuSelectHQ_Make(MissionMenu.MenuItem Menu, Player CurPlayer, object[] aExtraData)
    {
        foreach(CRadarHeadquarter HQ in m_aHeadQuarters)
        {
            if(HQ.GetOwningArmy() == CurPlayer.Army())
            {
                MissionMenu.MenuItem Command = Menu.AddDynamicCommand(HQ.GetName());
                Command.m_bEnabled = HQ.IsInRange(CurPlayer);
            }
        }
    }
    /// <summary>
    /// Delegate to process CMissionMenu's dynamic Menu commad selected.
    /// 
    /// * This is closely tied to OnMissionMenuSelectHQ_Make, as we only have the idx1Item (=the sequence) of the
    ///   menu's commands to connect to the selected command.
    ///   
    /// * ATTENTION: This might also be invoked if the command had been disabled (red color)
    /// 
    /// See CMissionMenu's explanation on Dynamic Menus for technical details.
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="idMenu"></param>
    /// <param name="idx1Item"></param>
    /// <returns>false if the menu command is/was actually meant to be disabled</returns>
    public bool OnMissionMenuSelectHQ_Select(Player CurPlayer, int idMenu, int idx1Item)
    {
        int idx = 0;
        foreach (CRadarHeadquarter HQ in m_aHeadQuarters)
        {
            if (HQ.GetOwningArmy() == CurPlayer.Army())
            {
                idx++;
                if (idx == idx1Item)
                {
                    if(!HQ.IsInRange(CurPlayer))
                    {
                        string sDenyReason = "Out of range";
                        m_Mission.GamePlay.gpHUDLogCenter(new Player[] { CurPlayer }, sDenyReason, null);
                        m_Mission.GamePlay.gpLogServer(new Player[] { CurPlayer }, sDenyReason, null);
                        return false;
                    }
                    CMissionPlayer MisPlayer = m_MissionPlayers.Get(CurPlayer);
                    MisPlayer.SetData(PLAYER_DATA_HQ, HQ.m_sStaticKey);
                    m_Messages.ConfirmConnectedToHeadquarter(CurPlayer, HQ);
                }
            }
        }
        return false;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="aExtraData"></param>
    public void OnMissionMenuGetEnemyContacts(Player CurPlayer, object[] aExtraData)
    {
        CRadarHeadquarter HQ = null;
        CMissionPlayer MisPlayer = null;

        if (!IsRequestValid(CurPlayer, out MisPlayer, out HQ))
        {
            return;
        }

        CPlayerRequest LastRequest = new CPlayerRequest();
        LastRequest.m_Time = DateTime.Now;

        CAircraftPulks Pulks = new CAircraftPulks(m_Mission);
        Pulks.LoadEnemies(CurPlayer, m_dPulkMaxRadius);
        Pulks.Filter(HQ, m_nMinAircraftInPulk, m_dMinObservableHeight);

        LastRequest.m_Report = m_Messages.CreateReport(Pulks, CurPlayer, false);
        MisPlayer.SetData(PLAYER_DATA_LASTREQUEST, LastRequest);
        m_Messages.SendReportOnEnemies(CurPlayer, HQ, LastRequest.m_Report, m_nResponseTime);
        PostReminderWhenWaitTimeIsOver(CurPlayer, HQ);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="aExtraData"></param>
    public void OnMissionMenuGetFriendlyContacts(Player CurPlayer, object[] aExtraData)
    {
        CRadarHeadquarter HQ = null;
        CMissionPlayer MisPlayer = null;

        if (!IsRequestValid(CurPlayer, out MisPlayer, out HQ))
        {
            return;
        }

        CPlayerRequest LastRequest = new CPlayerRequest();
        LastRequest.m_Time = DateTime.Now;

        CAircraftPulks Pulks = new CAircraftPulks(m_Mission);
        Pulks.LoadFriendlies(CurPlayer, m_dPulkMaxRadius);
        Pulks.Filter(HQ, m_nMinAircraftInPulk, m_dMinObservableHeight);

        LastRequest.m_Report = m_Messages.CreateReport(Pulks, CurPlayer, true);
        MisPlayer.SetData(PLAYER_DATA_LASTREQUEST, LastRequest);
        m_Messages.SendReportOnFriendlies(CurPlayer, HQ, LastRequest.m_Report, m_nResponseTime);
        PostReminderWhenWaitTimeIsOver(CurPlayer, HQ);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="missionNumber"></param>
    /// <param name="shortName"></param>
    /// <param name="active"></param>
    public CTargetDamage OnTrigger(int missionNumber, string shortName, bool active)
    {
        CTargetDamage td = null;
        if (shortName.StartsWith(TRIGGER_HQ))
        {
            for (int idxHQ = m_aHeadQuarters.Count - 1; idxHQ >= 0; idxHQ--)
            {
                if (shortName.Equals(m_aHeadQuarters[idxHQ].m_sTrigger))
                {
                    CRadarHeadquarter HQ = m_aHeadQuarters[idxHQ];
                    td = new CTargetDamage();
                    td.sName = HQ.GetName()+" ("+HQ.GetSector()+")";
                    td.iArmy = HQ.GetOwningArmy();
                    td.iPercent = 100;
                    td.bDestroyed = true;
                    td.iPoints = HQ.m_nPoints;

                    m_aHeadQuarters.RemoveAt(idxHQ);

                    return td;
                }
            }
        }
        else if (shortName.StartsWith(TRIGGER_OS))
        {
            for (int idxHQ = m_aHeadQuarters.Count - 1; idxHQ >= 0; idxHQ--)
            {
                CRadarHeadquarter HQ = m_aHeadQuarters[idxHQ];
                for (int idxOS = HQ.m_aObservers.Count - 1; idxOS >= 0; idxOS--)
                {
                    if (shortName.Equals(HQ.m_aObservers[idxOS].m_sTrigger))
                    {
                        CRadarObserver OS = HQ.m_aObservers[idxOS];
                        td = new CTargetDamage();
                        td.sName = OS.m_sName+" (" + OS.GetSector() + ")";
                        td.iArmy = HQ.GetOwningArmy();
                        td.iPercent = 100;
                        td.bDestroyed = true;
                        td.iPoints = 0;

                        HQ.m_aObservers.RemoveAt(idxOS);

                        /// if headquarter has no stations left, remove it.
                        if(0==HQ.m_aObservers.Count)
                        {
                            td.sName += ", last station of "+HQ.GetName() + ", ";
                            td.iPoints = HQ.m_nPoints;
                            m_aHeadQuarters.RemoveAt(idxHQ);
                        }

                        return td;
                    }
                }
            }
        }
        return td;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="nSeconds"></param>
    public void SetMinRequestInteval(int nSeconds)
    {
        m_nMinRequestInterval = nSeconds;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="nSeconds"></param>
    public void SetResponseTime(int nSeconds)
    {
        m_nResponseTime = nSeconds;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="dMeters"></param>
    public void SetMinObservableHeight(double dMeters)
    {
        m_dMinObservableHeight = dMeters;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="dRadius"></param>
    public void SetAircraftPulkRadius(double dRadius)
    {
        m_dPulkMaxRadius = dRadius;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="nAircraft"></param>
    public void SetMinAircraftInPulk(int nAircraft)
    {
        m_nMinAircraftInPulk = nAircraft;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="MisPlayer"></param>
    /// <param name="HQ"></param>
    /// <returns></returns>
    protected bool IsRequestValid(Player CurPlayer, out CMissionPlayer MisPlayer, out CRadarHeadquarter HQ)
    {
        MisPlayer = null;
        HQ = null;

        MisPlayer = m_MissionPlayers.Get(CurPlayer);
        if (null == MisPlayer)
        {
            return false;
        }
        /// get currently connected HQ
        string sStaticKey = (MisPlayer.GetData(PLAYER_DATA_HQ) as string);
        /// not connected
        if (null == sStaticKey)
        {
            m_Messages.InformThatPlayerNotConnected(CurPlayer);
            return false;
        }
        HQ = m_aHeadQuarters.Find(x => x.m_sStaticKey.Equals(sStaticKey));
        /// HQ does not exist anymore (destroyed)
        if (null == HQ)
        {
            m_Messages.InformThatHqNotActiveAnymore(CurPlayer);
            MisPlayer.RemoveData(PLAYER_DATA_HQ);
            MisPlayer.RemoveData(PLAYER_DATA_LASTREQUEST);
            return false;
        }
        /// out of range
        if (!HQ.IsInRange(CurPlayer))
        {
            m_Messages.InformThatPlayerOutOfRange(CurPlayer, HQ);
            MisPlayer.RemoveData(PLAYER_DATA_HQ);
            MisPlayer.RemoveData(PLAYER_DATA_LASTREQUEST);
            return false;
        }
        /// taken over by enemy
        if (HQ.GetOwningArmy() != CurPlayer.Army())
        {
            m_Messages.InformThatHQLostToEnemy(CurPlayer, HQ);
            MisPlayer.RemoveData(PLAYER_DATA_HQ);
            MisPlayer.RemoveData(PLAYER_DATA_LASTREQUEST);
            return false;
        }

        /// Previous request ... ? 
        CPlayerRequest LastRequest = MisPlayer.GetData(PLAYER_DATA_LASTREQUEST) as CPlayerRequest;
        /// ... Not imed out yet
        if (null != LastRequest && (DateTime.Now - LastRequest.m_Time).TotalSeconds < m_nMinRequestInterval)
        {
            m_Messages.AskToWaitAndRepeatPreviousReport(CurPlayer, LastRequest.m_Report, m_nMinRequestInterval);
            return false;
        }

        m_Messages.ConfirmRequest(CurPlayer, HQ);
        return true;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="CurPlayer"></param>
    /// <param name="HQ"></param>
    protected void PostReminderWhenWaitTimeIsOver(Player CurPlayer, CRadarHeadquarter HQ)
    {
        m_Mission.Timeout(m_nMinRequestInterval, () =>
        {
            m_Messages.InformThatWaitTimeIsOver(CurPlayer, HQ);
        });
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sMsg"></param>
    protected void LogMessage(string sMsg)
    {
        m_Mission.GamePlay.gpLogServer(null, "CRadar: " + sMsg, new object[] { });
    }
}

/*
class OriginalRadar
{
    LandMarkHandling LandMarks = new LandMarkHandling(
     new LandMark("Dover", 245577.51, 234521.20),
     new LandMark("Folkestone", 235333.17, 229568.64),
     new LandMark("St.Magarets's at Cliffe", 250423.81, 238001.64),
     new LandMark("Deal", 250985.14, 244801.80),
     new LandMark("Dymchurch", 223734.62, 222437.79),
     new LandMark("New Romney", 220005.71, 218110.40),
     new LandMark("Rye", 205758.66, 213525.22),
     new LandMark("Hastings", 195366.28, 203132.83),
     new LandMark("Redhill", 141166.70, 243753.00),
     new LandMark("Winchester", 61637.30, 222814.30),
     new LandMark("Guildford", 113168.50, 243068.00),
     new LandMark("Slough", 112057.30, 273232.60),
     new LandMark("Westminster", 143718.10, 272898.10),
     new LandMark("Dagenham", 162759.40, 277760.80),
     new LandMark("Tunbridge Wells", 172273.50, 233165.70),
     new LandMark("Maidstone", 189745.90, 249209.00),
     new LandMark("Canterbury", 228991.30, 251146.40),
     new LandMark("Ashford", 215339.80, 236055.90),
     new LandMark("Eastbourne", 174755.54, 191963.14),
     new LandMark("Brighton", 144909.86, 197927.40),
     new LandMark("Dungeness Lighthouse", 223253.40, 210155.30),
     new LandMark("Broadstairs", 253628.80, 261061.50),
     new LandMark("Selsey Bill", 99124.30, 185443.40),
     new LandMark("East Isle of Wight", 79054.00, 180916.70),
     new LandMark("Poole", 17546.20, 184304.90),
     new LandMark("Littlehampton", 117546.60, 195229.90),
     new LandMark("South Isle of Wight", 64160.80, 168414.00),
     new LandMark("West Isle of Wight", 45220.70, 179389.90),
     new LandMark("Sheerness", 206119.90, 268224.50),
     new LandMark("Herne Bay", 232080.30, 261816.80),
     new LandMark("Dunkirk", 320500.80, 225583.70),
     new LandMark("Ostend", 358023.90, 245730.30),
     new LandMark("St Omer", 312772.00, 192838.90),
     new LandMark("Boulogne-Sur-Mer", 266172.40, 189370.40),
     new LandMark("Cap Gris Nez", 264883.20, 206050.00),
     new LandMark("QuerqueVille", 33389.50, 70107.20),
     new LandMark("Tramecourt", 306241.20, 160780.20),
     new LandMark("Ste-Mere-Eglise", 55118.90, 43847.20),
     new LandMark("Caen", 125352.80, 19687.30),
     new LandMark("Le Havre", 160115.40, 53711.40),
     new LandMark("Dieppe", 229298.30, 101878.60),
     new LandMark("Abbeville", 282312.90, 121972.50),
     new LandMark("Le Touquet", 265264.30, 167753.30),
     new LandMark("St. Lo", 71599.20, 11523.50),
     new LandMark("Bayeux", 99913.50, 29508.90),
     new LandMark("Rouen", 229447.60, 47112.50),
     new LandMark("Les Andelys", 252889.60, 25790.50),
     new LandMark("Beauvais", 300777.90, 46676.70),
     new LandMark("Beaumont-Sur-Oise", 314635.80, 14481.90),
     new LandMark("Creil", 328713.80, 28098.30),
     new LandMark("Amiens", 316591.60, 97654.10),
     new LandMark("Roye", 351227.10, 76054.30),
     new LandMark("Arras", 349514.40, 141328.90),
     new LandMark("St. Pol-Sur-Ternoise", 318589.80, 151700.60),
     new LandMark("Leper", 356650.80, 204034.40),
     new LandMark("Bernay", 194302.50, 8880.50),
     new LandMark("Calais", 284654.87, 215707.54)
     );
    internal class LandMark
    {
        public string LandMarkName { get; set; }
        public Point2d LandMarkPosition { get; set; }

        public LandMark(string landMarkName, double x, double y)
        {
            this.LandMarkName = landMarkName;
            this.LandMarkPosition = new Point2d(x, y);
        }
    }
    internal class LandMarkHandling
    {
        List<LandMark> LandMarkList = new List<LandMark>();

        public LandMarkHandling(params LandMark[] mark)
        {
            if (mark != null)
                LandMarkList.AddRange(mark);
        }
        public LandMark getNearestLandMarkTo(Point3d position)
        {
            if (!(LandMarkList.Count > 0))
                return null;

            LandMark NearestLandMark = null;
            Point2d currentPosition = new Point2d(position.x, position.y);

            LandMarkList.ForEach(item =>
            {
                if (NearestLandMark != null)
                {
                    if (NearestLandMark.LandMarkPosition.distance(ref currentPosition) > item.LandMarkPosition.distance(ref currentPosition))
                        NearestLandMark = item;
                }
                else NearestLandMark = item;
            });

            return NearestLandMark;
        }
    }
    public void checkSectors(Player player)
    {
        foreach (var kvp in planePulks)
        {
            string sectorName = m_Mission.GamePlay.gpSectorName(kvp.Key.Pos().x, kvp.Key.Pos().y);

            if (localHQ.ObserveSector(sectorName) && kvp.Key.Pos().z > 400.00) //Separate out airgroups observed by the connected hq that are above 400.00 meters
            {
                string tmpKey = "";
                string message = "";
                Point3d destinationPoint = kvp.Key.Pos();
                LandMark nearestLandMark = LandMarks.getNearestLandMarkTo(kvp.Key.Pos());
                Point2d airgroupPos = new Point2d(kvp.Key.Pos().x, kvp.Key.Pos().y);
                Vector2d airgroupVector = new Vector2d(kvp.Key.Vwld().x, kvp.Key.Vwld().y);
                Point2d playerLocation = new Point2d(player.Place().Pos().x, player.Place().Pos().y);
                AiAircraft playerac = player.Place() as AiAircraft;
                AiAirGroup playerairgroup = playerac.AirGroup();
                Vector2d playerairgroupVector = new Vector2d(playerairgroup.Vwld().x, playerairgroup.Vwld().y);
                double playerspeed = 200;
                Point2d intercept = calculateInterceptionPoint(airgroupPos, airgroupVector, playerLocation, playerspeed);

                if (kvp.Value > 1) // only if more than 2 planes in a pulk generate a message
                {
                    foundEnemy = true;
                    if (nearestLandMark != null)
                    {
                        if (player.Army() != 1)
                        {
                            tmpKey = string.Format("INDIANER IN SECTOR {0}, {1} KILOMETERS {2} FROM {3} AT {4} METERS, HEADING {5} - VECTOR {6} TO INTERCEPT", sectorName, ToKilometers(nearestLandMark.LandMarkPosition.distance(ref airgroupPos)), degreesToWindRose(calculateBearingFromOrigin(airgroupPos, nearestLandMark.LandMarkPosition)), nearestLandMark.LandMarkName, ToMeters(kvp.Key.Pos().z), getDegreesIn10Step(calculateBearingDegree(airgroupVector)), getDegreesIn10Step(calculateBearingFromOrigin(calculateInterceptionPoint(airgroupPos, airgroupVector, playerLocation, playerspeed), playerLocation)));
                            message = string.Format("{0}+ INDIANER IN SECTOR {1}, {2} KILOMETERS {3} FROM {4} AT {5} METERS, HEADING {6} - VECTOR {7} TO INTERCEPT", noOfAircraft(kvp.Value), sectorName, ToKilometers(nearestLandMark.LandMarkPosition.distance(ref airgroupPos)), degreesToWindRose(calculateBearingFromOrigin(airgroupPos, nearestLandMark.LandMarkPosition)), nearestLandMark.LandMarkName, ToMeters(kvp.Key.Pos().z), getDegreesIn10Step(calculateBearingDegree(airgroupVector)), getDegreesIn10Step(calculateBearingFromOrigin(calculateInterceptionPoint(airgroupPos, airgroupVector, playerLocation, playerspeed), playerLocation)));
                        }
                        else
                        {
                            tmpKey = string.Format("BANDITS IN SECTOR {0}, {1} MILES {2} FROM {3} AT ANGELS {4}, HEADING {5} - VECTOR {6} TO INTERCEPT", sectorName, ToMiles(nearestLandMark.LandMarkPosition.distance(ref airgroupPos)), degreesToWindRose(calculateBearingFromOrigin(airgroupPos, nearestLandMark.LandMarkPosition)), nearestLandMark.LandMarkName, ToAngels(kvp.Key.Pos().z), getDegreesIn10Step(calculateBearingDegree(airgroupVector)), getDegreesIn10Step(calculateBearingFromOrigin(calculateInterceptionPoint(airgroupPos, airgroupVector, playerLocation, playerspeed), playerLocation)));
                            message = string.Format("{0}+ BANDITS IN SECTOR {1}, {2} MILES {3} FROM {4} AT ANGELS {5}, HEADING {6} - VECTOR {7} TO INTERCEPT", noOfAircraft(kvp.Value), sectorName, ToMiles(nearestLandMark.LandMarkPosition.distance(ref airgroupPos)), degreesToWindRose(calculateBearingFromOrigin(airgroupPos, nearestLandMark.LandMarkPosition)), nearestLandMark.LandMarkName, ToAngels(kvp.Key.Pos().z), getDegreesIn10Step(calculateBearingDegree(airgroupVector)), getDegreesIn10Step(calculateBearingFromOrigin(calculateInterceptionPoint(airgroupPos, airgroupVector, playerLocation, playerspeed), playerLocation)));
                        }
                    }
                }
            }
        }

        if (localHQ.GetLostObservedSectors().Count > 0)
        {
            string lostSectors = "";

            localHQ.GetLostObservedSectors().ForEach(item =>
            {
                lostSectors += item + " ";
            });

            if (player.Army() != 1)
            {
                lostSectors = lostSectors.TrimEnd(' ');
                m_Mission.GamePlay.gpLogServer(new[] { player }, "FREYA network has lost tracking for these sectors: {0}!", new[] { lostSectors });
            }
            else
            {
                lostSectors = lostSectors.TrimEnd(' ');
                m_Mission.GamePlay.gpLogServer(new[] { player }, "Chain Home is blind for these sectors: {0}!", new[] { lostSectors });
            }

            if (!foundEnemy)
                if (player.Army() != 1)
                {
                    m_Mission.GamePlay.gpLogServer(new[] { player }, "There are no enemy contacts to report in remaining FREYA coverage.", null);
                }
                else
                {
                    m_Mission.GamePlay.gpLogServer(new[] { player }, "I have no trade plotted in remaining RDF coverage.", null);
                }

        }
    }
}
*/