///
/// CSectionFile
///
/// @Author  Team Fusion
/// @Version 1.0.0
///
/// === Example integration into Mission ===
///
/// //$include <put-your-path-here>/CSectionFile-vX.y.z.cs
///
using System;
using System.Collections.Generic;
using System.IO;
using maddox.game;

/// <summary>
/// NOTE:
/// - Keys within sections are *not* unique!
/// </summary>
public class CSectionFile {
    /// <summary>
    /// 
    /// </summary>
    protected AMission m_Mission = null;

    /// <summary>
    /// 
    /// </summary>
    protected class CSectFile_Item {
        /// <summary>
        /// 
        /// </summary>
        public string sKey;
        /// <summary>
        /// 
        /// </summary>
        public string sVal;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sKey"></param>
        /// <param name="sVal"></param>
        public CSectFile_Item(string sKey, string sVal)
        {
            this.sKey = sKey;
            this.sVal = sVal;
        }
    }
    /// <summary>
    /// 
    /// </summary>
    protected class CSectFile_Section {
        /// <summary>
        /// 
        /// </summary>
        public string m_sName = null;
        /// <summary>
        /// Keys within sections are *not* unique!
        /// </summary>
        public List<CSectFile_Item> m_Items = null;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        public CSectFile_Section(string sName)
        {
            m_sName = sName;
            m_Items = new List<CSectFile_Item>();
        }
    }
    /// <summary>
    /// 
    /// </summary>
    Dictionary<string, CSectFile_Section> m_Sections = null;
    /// <summary>
    /// 
    /// </summary>
    /// <param name="Mission"></param>
    public CSectionFile(AMission Mission)
    {
        m_Mission = Mission;
        m_Sections  = new Dictionary<string, CSectFile_Section>();
    }
    /// <summary>
    /// Keys within sections are *not* unique! This returns true if there's at least *one*.
    /// </summary>
    /// <param name="sSection"></param>
    /// <param name="sKey"></param>
    /// <returns></returns>
    public bool Exist(string sSection, string sKey)
    {
        if(!m_Sections.ContainsKey(sSection))
        {
            return false;
        }
        return m_Sections[sSection].m_Items.Exists(x => x.sKey.Equals(sKey));

    }
    /// <summary>
    /// Keys within sections are *not* unique! This returns the first!
    /// </summary>
    /// <param name="sSection"></param>
    /// <param name="sKey"></param>
    /// <param name="sDefault"></param>
    /// <returns></returns>
    public string Get(string sSection, string sKey, string sDefault)
    {
        string sValue = null;
        if (m_Sections.ContainsKey(sSection))
        {
            CSectFile_Item Item = m_Sections[sSection].m_Items.Find(x => x.sKey.Equals(sKey));
            if(null != Item)
            {
                sValue = Item.sVal;
            }
        }
        return sValue;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sPathFile"></param>
    public void Load(string sPathFile)
    {
        Stream File = null;

        m_Sections.Clear();

        /// open the file (we let it throw if something wrong)
        File = new FileStream(sPathFile, FileMode.Open);

        CSectFile_Section CurSection = null;
        char[] acChars = new char[256];
        int iPosInStr = 0;
        int iEndOfStr = 0;

        using (var rFile = new StreamReader(File, System.Text.Encoding.UTF8))
        {
            while (true)
            {
                string line = rFile.ReadLine();
                if (line == null)
                {
                    break;
                }

                int nChars = line.Length;
                if (nChars == 0)
                {
                    continue;
                }
                if (acChars.Length < nChars)
                {
                    acChars = new char[nChars];
                }
                line.CopyTo(0, acChars, 0, nChars);
                iPosInStr = 0;
                iEndOfStr = nChars;
                RemoveComments(ref acChars, ref iPosInStr, ref iEndOfStr);
                if (0 < iEndOfStr - iPosInStr)
                {
                    // Next (or first) section "[section-name]"
                    if ((acChars[iPosInStr] == '[') && (acChars[iEndOfStr - 1] == ']'))
                    {
                        // if not empty []
                        if (2 < iEndOfStr - iPosInStr)
                        {
                            iPosInStr++;
                            iEndOfStr--;
                            CurSection = new CSectFile_Section(new string(acChars, iPosInStr, iEndOfStr - iPosInStr));
                            try
                            {
                                m_Sections.Add(CurSection.m_sName, CurSection);
                            }
                            catch(Exception e)
                            {
                                LogMessage("Add Section ["+ CurSection.m_sName + "]" + e.Message);
                            }
                        }
                    }
                    else if (CurSection != null)
                    {
                        string sKey = ReadVariable(ref acChars, ref iPosInStr, ref iEndOfStr);
                        if (sKey != null)
                        {
                            CurSection.m_Items.Add(new CSectFile_Item(sKey, ReadValue(ref acChars, ref iPosInStr, ref iEndOfStr)));
                        }
                    }
                }
            }
        }
        File.Close();
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="acChars"></param>
    /// <param name="iPosInStr"></param>
    /// <param name="iEndOfStr"></param>
    private void RemoveComments(ref char[] acChars, ref int iPosInStr, ref int iEndOfStr)
    {
        // remove preceeding spaces
        while (iPosInStr < iEndOfStr)
        {
            if (acChars[iPosInStr] <= ' ')
            {
                iPosInStr++;
            }
            else
            {
                break;
            }
        }
        /// shift iEndOfStr to where the comment starts
        int iIdx = iPosInStr;
        bool bComment = false;
        while (iIdx < iEndOfStr)
        {
            char ch = acChars[iIdx];
            if (ch == '\t')
            {
                acChars[iIdx] = ch = ' ';
            }
            if ((ch < ' ') || (ch == '#') || (ch == ';'))
            {
                iEndOfStr = iIdx;
                break;
            }
            if (ch == '/')
            {
                if (bComment)
                {
                    iEndOfStr = iIdx - 1;
                    break;
                }
                bComment = true;
            }
            else
            {
                bComment = false;
            }
            iIdx++;
        }

        // remove trailing spaces
        while (iEndOfStr > iPosInStr)
        {
            if (acChars[iEndOfStr - 1] <= ' ')
            {
                iEndOfStr--;
            }
            else
            {
                break;
            }
        }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="acChars"></param>
    /// <param name="iPosInStr"></param>
    /// <param name="iEndOfStr"></param>
    /// <returns></returns>
    private string ReadVariable(ref char[] acChars, ref int iPosInStr, ref int iEndOfStr)
    {
        string sReturn = null;
        int i = iPosInStr;
        if (acChars[i] == '"')
        {
            while (i < iEndOfStr)
            {
                if (acChars[i] != '"')
                {
                    i++;
                }
                else
                {
                    break;
                }
            }
            if ((i == iEndOfStr) || (i == iPosInStr))
            {
                i = iPosInStr;
            }
            else
            {
                sReturn = new string(acChars, iPosInStr + 1, i - iPosInStr - 2);
                iPosInStr = i + 1;
                return sReturn;
            }

        }
        while (i < iEndOfStr)
        {
            if (acChars[i] > ' ')
            {
                i++;
            }
            else
            {
                break;
            }
        }
        sReturn = new string(acChars, iPosInStr, i - iPosInStr);
        iPosInStr = i;
        return sReturn;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="acChars"></param>
    /// <param name="iPosInStr"></param>
    /// <param name="iEndOfStr"></param>
    /// <returns></returns>
    private string ReadValue(ref char[] acChars, ref int iPosInStr, ref int iEndOfStr)
    {
        while (iPosInStr < iEndOfStr)
        {
            if (acChars[iPosInStr] <= ' ')
            {
                iPosInStr++;
            }
            else
            {
                break;
            }
        }
        if (iPosInStr >= iEndOfStr)
        {
            return "";
        }
        return new string(acChars, iPosInStr, iEndOfStr - iPosInStr);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="sMsg"></param>
    protected void LogMessage(string sMsg)
    {
        m_Mission.GamePlay.gpLogServer(null, "CSectionFile: " + sMsg, new object[] { });
    }
}
