///
/// Manage Mission Targets: Convoys
///
/// @requires
///          CArmies-v1.0.1.cs
///          CTargetDamage-v1.0.1.cs
///
/// @Author    Team Fusion
/// @Version   1.0.0
///
/// === Example integration into Mission ===
///
/// //$include <put-your-path-here>/CTargetConvoys-vX.y.z.cs
///
///
///
using System;
using System.Collections;
using System.Collections.Generic;  /// List
using maddox.game;                 /// AMission
public class CTargetConvoy {
    public bool DEBUG_MESSAGES = false;
    /// Percentage at which the *convoy* is considered destroyed
    protected int m_iDestroyedAt = 100;
    /// the mission
    protected AMission m_Mission = null;
    
    protected int m_iArmy = CArmies.ARMY_NONE;
    protected string m_sName = "Unknown Convoy";
    protected int m_iPoints = 0;
    
    protected List<string> m_aTriggerNames = null;
    protected int m_nUnitsDestroyed = 0;
    /// <summary>
    /// Constructor
    /// </summary> 
    /// <param name="Mission">The Mission (use 'this')</param>
    /// <param name="iArmy">Army the airfield belongs to</param>
    /// <param name="sName">Name of the convoy</param>
    public CTargetConvoy(AMission Mission, int iArmy, string sName, int iPoints)
    {
        m_Mission = Mission;
        m_aTriggerNames = new List<string>();
        m_iArmy = iArmy;
        m_sName = sName;
        m_iPoints = iPoints;
    }
    /// <summary>
    /// Set percentage at which the convoy is considered destroyed
    /// </summary> 
    /// <param name="iPercent">Precentage</param>
    public void SetDestroyedAtPercent(int iPercent)
    {
        m_iDestroyedAt = iPercent;
    }
    /// <summary>
    /// </summary> 
    /// <param name=""></param>
    public void AddUnitByTriggerName(string sTriggerName)
    {
        m_aTriggerNames.Add(sTriggerName);
    }
    /// <summary>
    /// </summary> 
    /// <param name=""></param>
    public CTargetDamage OnTrigger(int missionNumber, string shortName, bool active)
    {
        foreach(string sTriggerName in m_aTriggerNames)
        {
            if (sTriggerName.Equals(shortName) && active)
            {
                m_Mission.GamePlay.gpGetTrigger(shortName).Enable = false;
                
                CTargetDamage Di = new CTargetDamage();
                double dOldDamage = m_nUnitsDestroyed;
                double dNewDamage = ++m_nUnitsDestroyed; 

                int iOldPercent  = Convert.ToInt32(Math.Ceiling((dOldDamage * 100)/(double)(m_aTriggerNames.Count)));
                int iNewPercent  = Convert.ToInt32(Math.Ceiling((dNewDamage * 100)/(double)(m_aTriggerNames.Count)));
                
                Di.sName      = m_sName;
                Di.iArmy      = m_iArmy;
                Di.iPercent   = iNewPercent;
                Di.bDestroyed = iNewPercent >= m_iDestroyedAt;
                Di.iPoints    = Di.bDestroyed ? m_iPoints : 0;
                
                return Di;
            }
        }
        return null;
    }
}
/// <summary>
/// </summary> 
public class CTargetConvoys {
    public bool DEBUG_MESSAGES = false;
    /// the mission
    protected AMission m_Mission = null;
    ///
    protected List<CTargetConvoy> m_aConvoys = null;
    /// <summary>
    /// Constructor
    /// </summary> 
    /// <param name="Mission">The Mission (use 'this')</param>
    public CTargetConvoys(AMission Mission)
    {
        m_Mission = Mission;
        m_aConvoys = new List<CTargetConvoy>();
    }
    /// <summary>
    /// </summary> 
    /// <param name=""></param>
    public void AddConvoy(CTargetConvoy Convoy)
    {
        m_aConvoys.Add(Convoy);
    }
    /// <summary>
    /// </summary> 
    /// <param name=""></param>
    public CTargetDamage OnTrigger(int missionNumber, string shortName, bool active)
    {
        foreach(CTargetConvoy Convoy in m_aConvoys)
        {
            CTargetDamage Di = Convoy.OnTrigger(missionNumber, shortName, active);
            if(null != Di)
            {
                return Di;
            }
        }
        return null;
    }
}