//$reference System.Data.dll
//$reference .\MySql.Data.dll
using System;
using System.IO;
using System.Text;
using System.Runtime.InteropServices;
using MySql.Data.MySqlClient;

using maddox.game;

namespace ServerStats {
    /// <summary>
    /// 
    /// </summary>
    public class CServerStats
    {
        /// <summary>
        /// 
        /// </summary>
        public bool m_bLogEvents = false;
        /// <summary>
        /// 
        /// also .ini
        /// </summary>
        protected bool m_bIsEnabled = false;
        /// <summary>
        /// Enable to get extensive log output
        /// also .ini
        /// </summary>
        protected bool m_bLogDbTraffic = false;
        /// <summary>
        /// 
        /// </summary>
        protected MySqlConnection m_conDb = null;
        /// <summary>
        /// 
        /// </summary>
        protected AMission m_Mission = null;
        /// <summary>
        /// 
        /// </summary>
        protected string m_sDbHost;
        /// <summary>
        /// 
        /// </summary>
        protected string m_sDbPort;
        /// <summary>
        /// 
        /// </summary>
        protected string m_sDbUsername;
        /// <summary>
        /// 
        /// </summary>
        protected string m_sDbPassword;
        /// <summary>
        /// 
        /// </summary>
        protected string m_sDbDatabase;
        /// <summary>
        /// 
        /// </summary>
        protected string m_sDbTableCurrentMissions;
        /// <summary>
        /// 
        /// </summary>
        protected string m_sDbTableCurrentPlayers;
        /// <summary>
        /// 
        /// </summary>
        protected string m_sDbTablePlayerStats;
        /// <summary>
        /// 
        /// </summary>
        protected string m_sGameserverName;
        /// <summary>
        /// 
        /// </summary>
        protected int m_iScoreDecimals = 2;
        /// <summary>
        /// 
        /// </summary>
        protected TimeSpan TimeLeftLast = new TimeSpan(0);
        /// <summary>
        /// In minutes
        /// </summary>
        protected int m_MinIntervalTimeLeftUpdate = 5;
        /// <summary>
        /// 
        /// </summary>
        /// <param name="Mission"></param>
        /// <param name="sCfgConnection"></param>
        public CServerStats(AMission Mission)
        {
            m_Mission = Mission;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public void StartStats(string sIniFile)
        {

            if(!File.Exists(sIniFile))
            {
                LogMessage("Start, configuration file not found: "+sIniFile);
                LogMessage("disabled");
                return;
            }

            CIniFile iniFile = new CIniFile(sIniFile);

            m_bIsEnabled = 0 != iniFile.Read("General", "IsEnabled", 0);

            m_sDbHost = iniFile.Read("Database", "Host", true);
            m_sDbPort = iniFile.Read("Database", "Port", true);
            m_sDbUsername = iniFile.Read("Database", "Username", true);
            m_sDbPassword = iniFile.Read("Database", "Password", true);
            m_sDbDatabase = iniFile.Read("Database", "Database", true);
            m_sDbTableCurrentMissions = iniFile.Read("Database", "TableCurrentMissions", true);
            m_sDbTableCurrentPlayers = iniFile.Read("Database", "TableCurrentPlayers", true);
            m_sDbTablePlayerStats = iniFile.Read("Database", "TablePlayerStats", true);

            m_sGameserverName = iniFile.Read("Gameserver", "Name", true);
            m_MinIntervalTimeLeftUpdate = Math.Max(1, iniFile.Read("Gameserver", "MinIntervalTimeLeftUpdate", 5));

            m_iScoreDecimals = iniFile.Read("Scores", "Decimals", m_iScoreDecimals);

            m_bLogDbTraffic = 0 != iniFile.Read("Debugging", "LogDbTraffic", 0);
            m_bLogEvents = 0 != iniFile.Read("Debugging", "LogEvent", 0);

            if(m_bIsEnabled)
            {
                if (m_bLogDbTraffic)
                {
                    LogMessage("m_sDbHost                : " + m_sDbHost                );
                    LogMessage("m_sDbPort                : " + m_sDbPort                );
                    LogMessage("m_sDbUsername            : " + m_sDbUsername            );
                    LogMessage("m_sDbPassword            : " + m_sDbPassword            );
                    LogMessage("m_sDbDatabase            : " + m_sDbDatabase            );
                    LogMessage("m_sDbTableCurrentMissions: " + m_sDbTableCurrentMissions);
                    LogMessage("m_sDbTableCurrentPlayers : " + m_sDbTableCurrentPlayers );
                    LogMessage("m_sDbTablePlayerStats    : " + m_sDbTablePlayerStats    );
                    LogMessage("m_sGameserverName        : " + m_sGameserverName        );
                    LogMessage("m_iScoreDecimals         : " + m_iScoreDecimals         );
                }
                if(!ConnectToDatabase())
                {
                    LogMessage("disabled");
                    return;
                }
                ClearCurrentMission();
                LogMessage("started");
            }
            else
            {
                LogMessage("disabled");
            }
        }
        /// <summary>
        /// 
        /// </summary>
        public void EndStats()
        {
            LogMessage("ended");
            ClearCurrentMission();
            DisconnectFromDatabase();
        }
        /// <summary>
        /// 
        /// </summary>
        public void ClearCurrentMission()
        {
            if (null == m_conDb)
            {
                return;
            }
            ClearCurrentMissionPlayers();
            if (m_bLogDbTraffic) LogMessage("ClearCurrentMission");
            MySqlCommand cmdSql = new MySqlCommand();
            cmdSql.CommandText = "DELETE FROM `" + m_sDbTableCurrentMissions + "` WHERE sServerName=@sServerName";
            cmdSql.Connection = m_conDb;
            cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
            ExecuteDatabaseNonQuery(cmdSql);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        /// <param name="sBriefing"></param>
        public void SetCurrentMission(string sName, string sBriefing)
        {
            if (null == m_conDb)
            {
                return;
            }
            if (m_bLogDbTraffic) LogMessage("SetCurrentMission");
            MySqlCommand cmdSql = new MySqlCommand();
            /// We set the TimeLeft to a default 4 hours - it might take some seconds before the first UpdateCurrentMissionTimeLeft kicks in ...
            cmdSql.CommandText = "REPLACE INTO `"+m_sDbTableCurrentMissions+"` SET sServerName=@sServerName, sName=@sName, sBriefing=@sBriefing, iTimeLeft=14400";
            cmdSql.Connection = m_conDb;
            cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
            cmdSql.Parameters.AddWithValue("@sName", sName);
            cmdSql.Parameters.AddWithValue("@sBriefing", sBriefing);
            ExecuteDatabaseNonQuery(cmdSql);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sPathBriefing">Relative to CloD-UserDoc</param>
        public void SetCurrentMissionFromBriefing(string sPathBriefing)
        {
            if (null == m_conDb)
            {
                return;
            }
            if (m_bLogDbTraffic) LogMessage("SetCurrentMissionFromBriefing");
            string sName = Path.GetFileNameWithoutExtension(sPathBriefing);
            string sAbsPath = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments) + @"\1C SoftClub\il-2 sturmovik cliffs of dover\" + sPathBriefing;
            string sBriefing = File.ReadAllText(sAbsPath);
            SetCurrentMission(sName, sBriefing);
            string sImageFile = Path.GetDirectoryName(sAbsPath) + @"\" + sName+".jpg";
            if(File.Exists(sImageFile))
            {
                UpdateCurrentMissionImage(sImageFile);
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sImageFile"></param>
        public void UpdateCurrentMissionImage(string sImageFile)
        {
            if (null == m_conDb)
            {
                return;
            }
            if (m_bLogDbTraffic) LogMessage("UpdateCurrentMissionImage: " + sImageFile);
            byte[] binImage = File.ReadAllBytes(sImageFile);
            MySqlCommand cmdSql = new MySqlCommand();
            cmdSql.CommandText = "UPDATE `" + m_sDbTableCurrentMissions + "` SET binImage=@binImage WHERE sServerName=@sServerName";
            cmdSql.Connection = m_conDb;
            cmdSql.Parameters.AddWithValue("@binImage", binImage);
            cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
            ExecuteDatabaseNonQuery(cmdSql);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="TimeLeft"></param>
        public void UpdateCurrentMissionTimeLeft(TimeSpan TimeLeft)
        {
            if (null == m_conDb)
            {
                return;
            }
            /// no more updates needed (also prevents TimeLeftLast becoming 0 again)
            if (0 >= TimeLeft.TotalMinutes || (TimeLeftLast.TotalSeconds != 0 && TimeLeftLast.TotalMinutes < m_MinIntervalTimeLeftUpdate))
            {
                return;
            }
            /// TimeLeftLast has been set at least once and minimum interval has not yet been exceeded
            if (TimeLeftLast.TotalSeconds != 0 && TimeLeftLast.TotalMinutes - m_MinIntervalTimeLeftUpdate < TimeLeft.TotalMinutes)
            {
                return;
            }
            TimeLeftLast = TimeLeft.Duration();
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("UpdateCurrentMissionTimeLeft: "+ TimeLeft.TotalSeconds);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "UPDATE `"+m_sDbTableCurrentMissions+"` SET iTimeLeft=@iTimeLeft WHERE sServerName=@sServerName";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@iTimeLeft", TimeLeftLast.TotalSeconds);
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        public void AddCurrentMissionPlayer(string sName)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("AddCurrentMissionPlayer " + sName);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "REPLACE INTO `" + m_sDbTableCurrentPlayers + "` SET sServerName=@sServerName, sName=@sName, sArmy='none', sVehicle='', sPosition=''";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        public void ClearCurrentMissionPlayers()
        {
            if (null == m_conDb)
            {
                return;
            }
            if (m_bLogDbTraffic) LogMessage("ClearCurrentMissionPlayers");
            MySqlCommand cmdSql = new MySqlCommand();
            cmdSql.CommandText = "DELETE FROM `" + m_sDbTableCurrentPlayers + "` WHERE sServerName=@sServerName";
            cmdSql.Connection = m_conDb;
            cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
            ExecuteDatabaseNonQuery(cmdSql);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        public void RemoveCurrentMissionPlayer(string sName)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("RemoveCurrentMissionPlayer " + sName);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "DELETE FROM `" + m_sDbTableCurrentPlayers + "` WHERE sServerName=@sServerName AND sName=@sName";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// Update sArmy, sVehicle, sPosition of a player.
        /// This adds the player if he's not already in table, because if a player was already connected
        /// on "battle start", we never got the OnPlayerConnected => AddCurrentMissionPlayer
        /// </summary>
        /// <param name="sName"></param>
        /// <param name="sArmy"></param>
        /// <param name="sVehicle"></param>
        /// <param name="sPosition"></param>
        public void UpdateCurrentMissionPlayerOnPlaceEnter(string sName, string sArmy, string sVehicle, string sPosition)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("UpdateCurrentMissionPlayerOnPlaceEnter '" + sName + "' '" + sArmy + "' '" + sVehicle + "' '" + sPosition + "'");
                MySqlCommand cmdSql = new MySqlCommand();
                /// We use INSERT ... ON DUPLICATE KEY UPDATE, because if a player was already connected
                /// on "battle start", we never got the OnPlayerConnected => AddCurrentMissionPlayer
                cmdSql.CommandText = "INSERT INTO `" + m_sDbTableCurrentPlayers + "` (sServerName, sName, sArmy, sVehicle, sPosition) VALUES (@sServerName, @sName, @sArmy, @sVehicle, @sPosition) ON DUPLICATE KEY UPDATE sArmy=VALUES(sArmy), sVehicle=VALUES(sVehicle), sPosition=VALUES(sPosition)";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                cmdSql.Parameters.AddWithValue("@sArmy", sArmy);
                cmdSql.Parameters.AddWithValue("@sVehicle", sVehicle);
                cmdSql.Parameters.AddWithValue("@sPosition", sPosition);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// Update sArmy, sVehicle of a player, sets position to ''.
        /// This only updates existing players, because if the player had disconnected while still in a
        /// vehicle onPlaceLeave is called *after* disconnect - we do not want to add him back in
        /// </summary>
        /// <param name="sName"></param>
        /// <param name="sArmy"></param>
        /// <param name="sVehicle"></param>
        public void UpdateCurrentMissionPlayerOnPlaceLeave(string sName, string sArmy, string sVehicle)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("UpdateCurrentMissionPlayerOnPlaceLeave '" + sName + "' '" + sArmy + "' '" + sVehicle + "'");
                MySqlCommand cmdSql = new MySqlCommand();
                /// We use UPDATE because if the player had disconnected (and onPlaceLeave is called *after* that)
                /// we do not want to add him back in
                cmdSql.CommandText = "UPDATE `" + m_sDbTableCurrentPlayers + "` SET sArmy=@sArmy, sVehicle=@sVehicle, sPosition='' WHERE sServerName=@sServerName AND sName=@sName";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                cmdSql.Parameters.AddWithValue("@sArmy", sArmy);
                cmdSql.Parameters.AddWithValue("@sVehicle", sVehicle);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        public void AddPlayerBailOut(string sName)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("AddPlayerBailOut " + sName);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "INSERT INTO `"+m_sDbTablePlayerStats+"` SET sServerName=@sServerName, sName=@sName, nBailOut=1 ON DUPLICATE KEY UPDATE nBailOut=nBailOut+1";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        public void AddPlayerCrashLanded(string sName)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("AddPlayerCrashLanded " + sName);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "INSERT INTO `"+m_sDbTablePlayerStats+"` SET sServerName=@sServerName, sName=@sName, nCrashLanding=1 ON DUPLICATE KEY UPDATE nCrashLanding=nCrashLanding+1";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        public void AddPlayerDeath(string sName)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("AddPlayerDeath " + sName);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "INSERT INTO `"+m_sDbTablePlayerStats+"` SET sServerName=@sServerName, sName=@sName, nDeath=1 ON DUPLICATE KEY UPDATE nDeath=nDeath+1";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        /// <param name="dKillAir"></param>
        public void AddPlayerKillAir(string sName, double dKillAir)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("AddPlayerKillAir " + sName + " "+ dKillAir);
                double dRounded = RoundScore(dKillAir);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "INSERT INTO `"+m_sDbTablePlayerStats+"` (sServerName, sName, dKillAir) VALUES (@sServerName, @sName, @dKillAir) ON DUPLICATE KEY UPDATE dKillAir=dKillAir+VALUES(dKillAir)";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                cmdSql.Parameters.AddWithValue("@dKillAir", dRounded);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        /// <param name="dKillGround"></param>
        public void AddPlayerKillGround(string sName, double dKillGround)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("AddPlayerKillGround " + sName + " " + dKillGround);
                double dRounded = RoundScore(dKillGround);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "INSERT INTO `"+m_sDbTablePlayerStats+"` (sServerName, sName, dKillGround) VALUES (@sServerName, @sName, @dKillGround) ON DUPLICATE KEY UPDATE dKillGround=dKillGround+VALUES(dKillGround)";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                cmdSql.Parameters.AddWithValue("@dKillGround", dRounded);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        /// <param name="dKillSea"></param>
        public void AddPlayerKillSea(string sName, double dKillSea)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("AddPlayerKillSea " + sName + " " + dKillSea);
                double dRounded = RoundScore(dKillSea);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "INSERT INTO `"+m_sDbTablePlayerStats+"` (sServerName, sName, dKillSea) VALUES (@sServerName, @sName, @dKillSea) ON DUPLICATE KEY UPDATE dKillSea=dKillSea+VALUES(dKillSea)";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                cmdSql.Parameters.AddWithValue("@dKillSea", dRounded);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        public void AddPlayerLanded(string sName)
        {
            if (null == m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("AddPlayerLanded " + sName);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "INSERT INTO `"+m_sDbTablePlayerStats+"` SET sServerName=@sServerName, sName=@sName, nLanding=1 ON DUPLICATE KEY UPDATE nLanding=nLanding+1";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sName"></param>
        public void AddPlayerTakeOff(string sName)
        {
            if(null==m_conDb)
            {
                return;
            }
            m_Mission.Timeout(0.1, () => {
                if (m_bLogDbTraffic) LogMessage("AddPlayerTakeOff "+sName);
                MySqlCommand cmdSql = new MySqlCommand();
                cmdSql.CommandText = "INSERT INTO `"+m_sDbTablePlayerStats+"` SET sServerName=@sServerName, sName=@sName, nTakeOff=1 ON DUPLICATE KEY UPDATE nTakeOff=nTakeOff+1";
                cmdSql.Connection = m_conDb;
                cmdSql.Parameters.AddWithValue("@sServerName", m_sGameserverName);
                cmdSql.Parameters.AddWithValue("@sName", sName);
                ExecuteDatabaseNonQuery(cmdSql);
            });
        }
        /// <summary>
        /// Establish connection to the database
        /// </summary>
        protected bool ConnectToDatabase()
        {
            string sCfgConnection = 
                 "Host="+m_sDbHost
                +";Port="+m_sDbPort
                +";Username="+m_sDbUsername
                +";Password="+m_sDbPassword
                +";Database="+m_sDbDatabase
            ;
            try
            {
                m_conDb = new MySqlConnection(sCfgConnection);
                m_conDb.Open();
                if (m_bLogDbTraffic)
                {
                    LogMessage("MySqlConnection.Open " + sCfgConnection);
                }
                return true;
            }
            catch (MySql.Data.MySqlClient.MySqlException e)
            {
                m_conDb = null;
                switch (e.Number)
                {
                    case 0:
                        LogMessage("Cannot connect to server(" + sCfgConnection + ")");
                        break;
                    case 1045:
                        LogMessage("Invalid username/password, please try again");
                        break;
                    default:
                        LogMessage("ERROR MySqlConnection.Open(" + sCfgConnection + "): " + e.Message);
                        break;
                }
            }
            catch (Exception e)
            {
                m_conDb = null;
                LogMessage("ERROR MySqlConnection.Open("+ sCfgConnection + "): " + e.Message);
            }
            return false;
        }
        /// <summary>
        /// 
        /// </summary>
        protected void DisconnectFromDatabase()
        {
            if (null != m_conDb)
            {
                try
                {
                    m_conDb.Close();
                    if (m_bLogDbTraffic)
                    {
                        LogMessage("MySqlConnection.Close");
                    }
                }
                catch (Exception e)
                {
                    LogMessage("ERROR MySqlConnection.Close: " + e.Message);
                }
                m_conDb = null;
            }
        }
        /// <summary>
        /// Execute a non-query and catch exceptions
        /// </summary>
        /// <param name="cmdSql"></param>
        protected void ExecuteDatabaseNonQuery(MySqlCommand cmdSql)
        {
            /// in case we've lost connection to database server, we try to reconnect
            if(cmdSql.Connection.State == System.Data.ConnectionState.Closed)
            {
                try
                {
                    cmdSql.Connection.Open();
                }
                catch(Exception e)
                {
                    if (m_bLogDbTraffic)
                    {
                        LogMessage("ExecuteNonQuery, failure trying to reconnect closed connection: " + e.Message);
                    }
                    return;
                }
            }
            try
            {
                cmdSql.ExecuteNonQuery();
            }
            catch (Exception e)
            {
                LogMessage("ExecuteNonQuery Exception: " + e.Message);
                if (m_bLogDbTraffic)
                {
                    LogMessage("ExecuteNonQuery   - ConState: " + cmdSql.Connection.State.ToString());
                    LogMessage("ExecuteNonQuery   - Command : " + cmdSql.CommandText);
                }
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="score0_100"></param>
        /// <returns></returns>
        protected double RoundScore(double score0_100)
        {
            return Math.Round(score0_100 / 100, m_iScoreDecimals, MidpointRounding.AwayFromZero);
        }
        /// <summary>
        /// Read values from a .ini file
        /// </summary>
        protected class CIniFile   // revision 11
        {
            readonly string m_sPath;

            [DllImport("kernel32", CharSet = CharSet.Unicode)]
            static extern int GetPrivateProfileString(string Section, string Key, string Default, StringBuilder RetVal, int Size, string FilePath);

            public CIniFile(string IniPath)
            {
                m_sPath = new FileInfo(IniPath).FullName;
            }
            public string Read(string Section, string Key, bool bThrowOnFailure)
            {
                var RetVal = new StringBuilder(255);
                GetPrivateProfileString(Section, Key, "", RetVal, 255, m_sPath);
                if(0==RetVal.Length && bThrowOnFailure)
                {
                    throw new Exception("ini value [" + Section + "] '" + Key + "' not found!");
                }
                return RetVal.ToString();
            }
            public string Read(string Section, string Key, string sDefault)
            {
                string sVal = Read(Section, Key, false);
                if (sVal.Length < 0)
                {
                    sVal = sDefault;
                }
                return sVal;
            }
            public int Read(string Section, string Key, int iDefault)
            {
                int iVal;
                string sVal = Read(Section, Key, false);
                if(0 < sVal.Length)
                {
                    iVal = int.Parse(sVal);
                }
                else
                {
                    iVal = iDefault;
                }
                return iVal;
            }
            public bool KeyExists(string Section, string Key)
            {
                return Read(Section, Key, false).Length > 0;
            }
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="sMsg"></param>
        protected void LogMessage(string sMsg)
        {
            m_Mission.GamePlay.gpLogServer(null, "CServerStats: " + sMsg, new object[] { });
        }
    }
} /// end namespace